from pathlib import Path

import fpdf
from fpdf.errors import FPDFException
from test.conftest import assert_pdf_equal, assert_same_file

import pytest

HERE = Path(__file__).resolve().parent
POLYLINE_COORDINATES = [(10, 10), (40, 10), (10, 40)]
SCALING_FACTORS_FOR_UNITS = [
    ("pt", 1),
    ("mm", 1 / (72 / 25.4)),
    ("cm", 1 / (72 / 2.54)),
    ("in", 1 / 72),
]


@pytest.mark.parametrize("unit, factor", SCALING_FACTORS_FOR_UNITS)
def test_polyline_command_all_k(unit, factor):
    pdf = fpdf.FPDF(unit=unit)
    pdf.add_page()
    data = []
    # pylint: disable=protected-access
    pdf._out = data.append

    pdf.polyline(scale_points(POLYLINE_COORDINATES, factor))
    assert "".join(data) == "10.00 831.89 m40.00 831.89 l10.00 801.89 l S"
    data.clear()

    with pytest.warns(DeprecationWarning) as record:
        # keep one "fill=True" to test the warning
        pdf.polyline(scale_points(POLYLINE_COORDINATES, factor), fill=True)
        assert "".join(data) == "10.00 831.89 m40.00 831.89 l10.00 801.89 l B"
        data.clear()

    assert len(record) == 1
    assert_same_file(record[0].filename, __file__)

    pdf.polyline(scale_points(POLYLINE_COORDINATES, factor), polygon=True)
    assert "".join(data) == "10.00 831.89 m40.00 831.89 l10.00 801.89 l h S"
    data.clear()

    pdf.polyline(scale_points(POLYLINE_COORDINATES, factor), polygon=True, style="DF")
    assert "".join(data) == "10.00 831.89 m40.00 831.89 l10.00 801.89 l h B"


def scale_points(raw_points, k_recip):
    return [(k_recip * coord[0], k_recip * coord[1]) for coord in raw_points]


def test_check_page():
    pdf = fpdf.FPDF(unit="pt")

    with pytest.raises(FPDFException) as polyline_no_page:
        pdf.polyline(POLYLINE_COORDINATES)
    expected_error_msg = "No page open, you need to call add_page() first"
    assert expected_error_msg == str(polyline_no_page.value)

    with pytest.raises(FPDFException) as polygon_no_page:
        pdf.polygon(POLYLINE_COORDINATES)
    assert expected_error_msg == str(polygon_no_page.value)


def test_filled_polygon(tmp_path):
    pdf = fpdf.FPDF()
    pdf.add_page()
    pdf.set_line_width(2)
    pdf.set_fill_color(r=255, g=0, b=0)
    coords = ((100, 0), (5, 69), (41, 181), (159, 181), (195, 69))
    pdf.polygon(coords, style="DF")
    assert_pdf_equal(pdf, HERE / "filled_polygon.pdf", tmp_path)



"""
Usage documentation at: <https://py-pdf.github.io/fpdf2/Presentations.html#transitions>
"""

from abc import ABC
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from .encryption import StandardSecurityHandler


class Transition(ABC):
    def serialize(
        self,
        _security_handler: Optional["StandardSecurityHandler"] = None,
        _obj_id: Optional[int] = None,
    ) -> str:
        raise NotImplementedError


class SplitTransition(Transition):
    def __init__(self, dimension: str, direction: str) -> None:
        if dimension not in ("H", "V"):
            raise ValueError(
                f"Unsupported dimension '{dimension}', must be H(horizontal) or V(ertical)"
            )
        self.dimension = dimension
        if direction not in ("I", "O"):
            raise ValueError(
                f"Unsupported direction '{direction}', must be I(nward) or O(utward)"
            )
        self.direction = direction

    def serialize(
        self,
        _security_handler: Optional["StandardSecurityHandler"] = None,
        _obj_id: Optional[int] = None,
    ) -> str:
        return f"<</Type /Trans /S /Split /DM /{self.dimension} /M /{self.direction}>>"


class BlindsTransition(Transition):
    def __init__(self, dimension: str) -> None:
        if dimension not in ("H", "V"):
            raise ValueError(
                f"Unsupported dimension '{dimension}', must be H(horizontal) or V(ertical)"
            )
        self.dimension = dimension

    def serialize(
        self,
        _security_handler: Optional["StandardSecurityHandler"] = None,
        _obj_id: Optional[int] = None,
    ) -> str:
        return f"<</Type /Trans /S /Blinds /DM /{self.dimension}>>"


class BoxTransition(Transition):
    def __init__(self, direction: str) -> None:
        if direction not in ("I", "O"):
            raise ValueError(
                f"Unsupported direction '{direction}', must be I(nward) or O(utward)"
            )
        self.direction = direction

    def serialize(
        self,
        _security_handler: Optional["StandardSecurityHandler"] = None,
        _obj_id: Optional[int] = None,
    ) -> str:
        return f"<</Type /Trans /S /Blinds /M /{self.direction}>>"


class WipeTransition(Transition):
    def __init__(self, direction: int) -> None:
        if direction not in (0, 90, 180, 270):
            raise ValueError(
                f"Unsupported direction '{direction}', must 0, 90, 180 or 270"
            )
        self.direction = direction

    def serialize(
        self,
        _security_handler: Optional["StandardSecurityHandler"] = None,
        _obj_id: Optional[int] = None,
    ) -> str:
        return f"<</Type /Trans /S /Wipe /Di /{self.direction}>>"


class DissolveTransition(Transition):
    def serialize(
        self,
        _security_handler: Optional["StandardSecurityHandler"] = None,
        _obj_id: Optional[int] = None,
    ) -> str:
        return "<</Type /Trans /S /Dissolve>>"


class GlitterTransition(Transition):
    def __init__(self, direction: int) -> None:
        if direction not in (0, 270, 315):
            raise ValueError(f"Unsupported direction '{direction}', must 0, 270 or 315")
        self.direction = direction

    def serialize(
        self,
        _security_handler: Optional["StandardSecurityHandler"] = None,
        _obj_id: Optional[int] = None,
    ) -> str:
        return f"<</Type /Trans /S /Glitter /Di /{self.direction}>>"


class FlyTransition(Transition):
    def __init__(self, dimension: str, direction: Optional[int] = None) -> None:
        if dimension not in ("H", "V"):
            raise ValueError(
                f"Unsupported dimension '{dimension}', must be H(horizontal) or V(ertical)"
            )
        self.dimension = dimension
        if direction not in (0, 270, None):
            raise ValueError(
                f"Unsupported direction '{direction}', must 0, 270 or None"
            )
        self.direction = direction

    def serialize(
        self,
        _security_handler: Optional["StandardSecurityHandler"] = None,
        _obj_id: Optional[int] = None,
    ) -> str:
        return (
            f"<</Type /Trans /S /Glitter /M /{self.dimension} /Di /{self.direction}>>"
        )


class PushTransition(Transition):
    def __init__(self, direction: int) -> None:
        if direction not in (0, 270):
            raise ValueError(f"Unsupported direction '{direction}', must 0 or 270")
        self.direction = direction

    def serialize(
        self,
        _security_handler: Optional["StandardSecurityHandler"] = None,
        _obj_id: Optional[int] = None,
    ) -> str:
        return f"<</Type /Trans /S /Push /Di /{self.direction}>>"


class CoverTransition(Transition):
    def __init__(self, direction: int) -> None:
        if direction not in (0, 270):
            raise ValueError(f"Unsupported direction '{direction}', must 0 or 270")
        self.direction = direction

    def serialize(
        self,
        _security_handler: Optional["StandardSecurityHandler"] = None,
        _obj_id: Optional[int] = None,
    ) -> str:
        return f"<</Type /Trans /S /Cover /Di /{self.direction}>>"


class UncoverTransition(Transition):
    def __init__(self, direction: int) -> None:
        if direction not in (0, 270):
            raise ValueError(f"Unsupported direction '{direction}', must 0 or 270")
        self.direction = direction

    def serialize(
        self,
        _security_handler: Optional["StandardSecurityHandler"] = None,
        _obj_id: Optional[int] = None,
    ) -> str:
        return f"<</Type /Trans /S /Uncover /Di /{self.direction}>>"


class FadeTransition(Transition):
    def serialize(
        self,
        _security_handler: Optional["StandardSecurityHandler"] = None,
        _obj_id: Optional[int] = None,
    ) -> str:
        return "<</Type /Fade /S /Dissolve>>"



from pathlib import Path

import pytest

from fpdf import FPDF
from fpdf.errors import FPDFException
from fpdf.enums import PageLayout, PageMode
from test.conftest import assert_pdf_equal

HERE = Path(__file__).resolve().parent


@pytest.mark.parametrize("zoom", ["fullpage", "fullwidth", "real", "default"])
def test_setting_all_zoom(zoom, tmp_path):
    """This test executes some possible inputs to FPDF#set_display_mode."""
    doc = FPDF()
    doc.add_page()
    doc.set_font("helvetica", size=12)
    doc.cell(
        w=72,
        h=0,
        border=1,
        text="hello world",
        new_x="LMARGIN",
        new_y="NEXT",
    )
    doc.set_display_mode(zoom=zoom)
    assert_pdf_equal(doc, HERE / f"zoom-{zoom}.pdf", tmp_path)


def test_setting_zoom_raises_correct_error():
    doc = FPDF()
    with pytest.raises(FPDFException):
        doc.set_display_mode("foo")


@pytest.mark.parametrize("page_layout", PageLayout.__members__.keys())
def test_page_layout(page_layout, tmp_path):
    pdf = FPDF()
    pdf.set_display_mode(zoom="default", layout=page_layout)
    pdf.set_font("helvetica", size=30)
    pdf.add_page()
    pdf.cell(text="page 1")
    pdf.add_page()
    pdf.cell(text="page 2")
    assert_pdf_equal(pdf, HERE / f"page-layout-{page_layout}.pdf", tmp_path)


@pytest.mark.parametrize("layout_alias", ["single", "continuous", "two", "default"])
def test_layout_aliases(layout_alias, tmp_path):
    pdf = FPDF()
    pdf.set_display_mode(zoom="default", layout=layout_alias)
    pdf.set_font("helvetica", size=30)
    pdf.add_page()
    pdf.cell(text="page 1")
    pdf.add_page()
    pdf.cell(text="page 2")
    assert_pdf_equal(pdf, HERE / f"layout-alias-{layout_alias}.pdf", tmp_path)


@pytest.mark.parametrize("page_mode", PageMode.__members__.keys())
def test_page_mode(page_mode, tmp_path):
    pdf = FPDF()
    pdf.page_mode = page_mode
    pdf.set_font("helvetica", size=30)
    pdf.add_page()
    pdf.cell(text="page 1")
    pdf.add_page()
    pdf.cell(text="page 2")
    assert_pdf_equal(pdf, HERE / f"page-mode-{page_mode}.pdf", tmp_path)


def test_page_mode_zoom_numeric(tmp_path):
    pdf = FPDF()
    pdf.set_display_mode(zoom=25)
    pdf.set_font("helvetica", size=30)
    pdf.add_page()
    pdf.cell(text="page 1")
    pdf.add_page()
    pdf.cell(text="page 2")
    assert_pdf_equal(
        pdf,
        HERE / "page-mode-zoom_numeric.pdf",
        tmp_path,
    )



# Configure mkdocs logging
# Inspired by: https://github.com/mkdocs/mkdocs/issues/2876
import logging

from mkdocs import plugins


@plugins.event_priority(100)
def on_startup(*_, **__):
    logger = logging.getLogger("mkdocs")
    # Instance of ColorFormatter defined in mkdocs/__main__.py:
    formatter = logger.handlers[0].formatter
    formatter.datefmt = "%H:%M:%S"
    formatter._style = logging.PercentStyle("%(asctime)s [%(name)s] %(message)s")



"""
Generate a full requirements file for GuardDog from pyproject.toml.
"""

import sys
from pathlib import Path
import tomllib


def main(pyproject_path: str, out_path: str) -> int:
    pyproject_file = Path(pyproject_path)
    if not pyproject_file.is_file():
        raise RuntimeError(f"pyproject.toml not found at {pyproject_file}")

    data = tomllib.loads(pyproject_file.read_text(encoding="utf-8"))

    project = data.get("project", {})
    base_deps = project.get("dependencies", [])
    extras = project.get("optional-dependencies", {})

    # Which extras to include in the "full" file
    extra_order = ["dev", "docs", "test"]

    combined: list[str] = []

    def add_deps(seq: list[str]) -> None:
        for dep in seq:
            if dep not in combined:
                combined.append(dep)

    add_deps(base_deps)

    for extra_name in extra_order:
        add_deps(extras.get(extra_name, []))

    out_file = Path(out_path)
    out_file.write_text("\n".join(combined) + "\n", encoding="utf-8")
    print(f"Wrote {len(combined)} dependencies to {out_file}")


if __name__ == "__main__":
    pyproject_arg = sys.argv[1] if len(sys.argv) > 1 else "pyproject.toml"
    out_arg = sys.argv[2] if len(sys.argv) > 2 else "fpdf2-full-requirements.txt"
    main(pyproject_arg, out_arg)



from fpdf import FPDF

pdf = FPDF()

# First page:
pdf.add_page()
pdf.set_font("helvetica", size=20)
pdf.write(5, "To find out what's new in self tutorial, click ")
pdf.set_font(style="U")
link = pdf.add_link(page=2)
pdf.write(5, "here", link)
pdf.set_font()

# Second page:
pdf.add_page()
pdf.image(
    "../docs/fpdf2-logo.png", 10, 10, 50, 0, "", "https://py-pdf.github.io/fpdf2/"
)
pdf.set_left_margin(60)
pdf.set_font_size(18)
pdf.write_html(
    """You can print text mixing different styles using HTML tags: <b>bold</b>, <i>italic</i>,
<u>underlined</u>, or <b><i><u>all at once</u></i></b>!
<br><br>You can also insert links on text, such as <a href="https://py-pdf.github.io/fpdf2/">https://py-pdf.github.io/fpdf2/</a>,
or on an image: the logo is clickable!"""
)
pdf.output("tuto6.pdf")



import logging
from pathlib import Path

from fpdf import FPDF
from test.conftest import assert_pdf_equal

HERE = Path(__file__).resolve().parent


def test_insert_png_files(caplog, tmp_path):
    with caplog.at_level(logging.INFO):
        pdf = FPDF()
        for png_path in sorted(HERE.glob("*.png")):
            pdf.add_page()
            pdf.image(png_path, x=0, y=0, w=0, h=0)
    # Note: 7 of those images have an ICC profile, and there are only 5 distinct ICC profiles among them
    assert_pdf_equal(pdf, HERE / "image_png_insert_png_files.pdf", tmp_path)

    assert "Unsupported color space CMYK in ICC Profile of file" in caplog.text
    # Note: the warning above comes from the following files, for which ImageMagics also raise warnings:
    #   identify-im6.q16: iCCP: profile 'icc': 'CMYK': invalid ICC profile color space `test/image/png_images/0839d93f8e77e21acd0ac40a80b14b7b.png'
    #   identify-im6.q16: iCCP: profile 'icc': 'CMYK': invalid ICC profile color space `test/image/png_images/1ebd73c1d3fbc89782f29507364128fc.png'



import pytest

from fpdf.errors import FPDFPageFormatException
from fpdf.fpdf import get_page_format


def test_page_format_ok():
    assert get_page_format("a4") == (595.28, 841.89)
    assert get_page_format("letter") == (612, 792)
    assert get_page_format((297, 210), k=2) == (594, 420)


def test_page_format_error():
    with pytest.raises(FPDFPageFormatException) as error:
        get_page_format("letter1")

    assert "FPDFPageFormatException" in str(error.value)
    assert "Unknown page format" in str(error.value)
    assert "letter1" in str(error.value)

    with pytest.raises(FPDFPageFormatException) as error:
        get_page_format(3)

    assert "FPDFPageFormatException" in str(error.value)
    assert "Only one argument given" in str(error.value)

    with pytest.raises(FPDFPageFormatException) as error:
        get_page_format(4, "a")

    assert "FPDFPageFormatException" in str(error.value)
    assert "Arguments must be numbers: " in str(error.value)



from pathlib import Path
from fpdf import FPDF
from fpdf.pattern import RadialGradient
from test.conftest import assert_pdf_equal

HERE = Path(__file__).resolve().parent


def test_radial_gradient(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    x = pdf.w / 2 - 25
    y = pdf.get_y()
    with pdf.use_pattern(
        RadialGradient(
            x + 25,
            y + 25,
            0,
            x + 25,
            y + 25,
            25,
            [(255, 255, 0), (255, 0, 0)],
        )
    ):
        pdf.circle(x=x + 25, y=y + 25, radius=25, style="FD")
    y += 60
    with pdf.use_pattern(
        RadialGradient(
            x + 5,
            y + 5,
            0,
            x + 25,
            y + 25,
            25,
            [(255, 255, 0), (255, 0, 0)],
        )
    ):
        pdf.circle(x=x + 25, y=y + 25, radius=25, style="FD")

    assert_pdf_equal(pdf, HERE / "radial_gradient.pdf", tmp_path)


def test_radial_gradient_multiple_colors(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    x = pdf.l_margin
    y = pdf.get_y()
    with pdf.use_pattern(
        RadialGradient(
            (pdf.epw + pdf.l_margin) / 2,
            y + 50,
            20,
            (pdf.epw + pdf.l_margin) / 2,
            y + 50,
            (pdf.epw + pdf.l_margin) / 2,
            ["#868F96", "#596164", "#537895", "#09203F"],
        )
    ):
        pdf.rect(x=x, y=y, w=pdf.epw, h=100, style="FD")
    y += 105

    with pdf.use_pattern(
        RadialGradient(
            pdf.w / 2,
            y + 50,
            0,
            pdf.w / 2,
            y + 50,
            (y + 50) / 2,
            ["#FFECD2", "#FCB69F", "#DD2476"],
        )
    ):
        pdf.rect(x=x, y=y, w=pdf.epw, h=100, style="FD")

    assert_pdf_equal(pdf, HERE / "radial_gradient_multiple_colors.pdf", tmp_path)


def test_custom_bounds(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    with pdf.use_pattern(
        RadialGradient(
            pdf.w / 2,
            pdf.h / 2,
            20,
            pdf.w / 2,
            pdf.h / 2,
            pdf.h / 2 - 20,
            [
                "#FFFFFF",
                "#9400D3",
                "#4B0082",
                "#0000FF",
                "#00FF00",
                "#FFFF00",
                "#FF7F00",
                "#FF0000",
                "#FFFFFF",
            ],
            bounds=[0.35, 0.40, 0.45, 0.50, 0.55, 0.60, 0.65],
        )
    ):
        pdf.rect(x=0, y=0, w=pdf.w, h=pdf.h / 2, style="FD")

    assert_pdf_equal(pdf, HERE / "radial_custom_bounds.pdf", tmp_path)



from pathlib import Path

import fpdf
import pytest
from test.conftest import assert_pdf_equal

HERE = Path(__file__).resolve().parent


def next_row(pdf):
    pdf.ln()
    pdf.set_y(pdf.get_y() + size + margin)


size = 50
margin = 10


def test_rect_not_square(tmp_path):
    pdf = fpdf.FPDF(unit="mm")
    pdf.add_page()

    for counter, style in enumerate(["", "F", "FD", "DF", None]):
        pdf.rect(x=pdf.get_x(), y=pdf.get_y(), w=size / 2, h=size, style=style)
        pdf.set_x(pdf.get_x() + (size / 2) + margin)
        if counter % 3 == 0:
            next_row(pdf)

    assert_pdf_equal(pdf, HERE / "class_rect_not_square.pdf", tmp_path)


def test_rect_style(tmp_path):
    pdf = fpdf.FPDF(unit="mm")
    pdf.add_page()

    for counter, style in enumerate(["", "F", "FD", "DF", None]):
        pdf.rect(x=pdf.get_x(), y=pdf.get_y(), w=size, h=size, style=style)
        pdf.set_x(pdf.get_x() + size + margin)
        if counter % 3 == 0:
            next_row(pdf)

    assert_pdf_equal(pdf, HERE / "class_rect_style.pdf", tmp_path)


def test_rect_line_width(tmp_path):
    pdf = fpdf.FPDF(unit="mm")
    pdf.add_page()

    for line_width in [1, 2, 3]:
        pdf.set_line_width(line_width)
        pdf.rect(x=pdf.get_x(), y=pdf.get_y(), w=size, h=size, style=None)
        pdf.set_x(pdf.get_x() + size + margin)
    next_row(pdf)
    for line_width in [4, 5, 6]:
        pdf.set_line_width(line_width)
        pdf.rect(x=pdf.get_x(), y=pdf.get_y(), w=size, h=size, style=None)
        pdf.set_x(pdf.get_x() + size + margin)
    pdf.set_line_width(0.2)  # reset

    assert_pdf_equal(pdf, HERE / "class_rect_line_width.pdf", tmp_path)


def test_rect_draw_color(tmp_path):
    pdf = fpdf.FPDF(unit="mm")
    pdf.add_page()

    # Colors
    pdf.set_line_width(0.5)
    for gray in [70, 140, 210]:
        pdf.set_draw_color(gray)
        pdf.rect(x=pdf.get_x(), y=pdf.get_y(), w=size, h=size, style=None)
        pdf.set_x(pdf.get_x() + size + margin)

    assert_pdf_equal(pdf, HERE / "class_rect_draw_color.pdf", tmp_path)


def test_rect_fill_color(tmp_path):
    pdf = fpdf.FPDF(unit="mm")
    pdf.add_page()

    pdf.set_fill_color(240)
    for color in [[230, 30, 180], [30, 180, 30], [30, 30, 70]]:
        pdf.set_draw_color(*color)
        pdf.rect(x=pdf.get_x(), y=pdf.get_y(), w=size, h=size, style="FD")
        pdf.set_x(pdf.get_x() + size + margin)

    next_row(pdf)

    assert_pdf_equal(pdf, HERE / "class_rect_fill_color.pdf", tmp_path)


def test_round_corners_rect(tmp_path):
    pdf = fpdf.FPDF()
    pdf.add_page()

    # no fill different sizes test
    y = 10
    pdf.rect(10, y, 15, 20, round_corners=True, style="D")
    pdf.rect(30, y, 20, 15, round_corners=True, style="D")
    pdf.rect(60, y, 33, 28, round_corners=True, style="D")
    pdf.rect(100, y, 50, 10, round_corners=True, style="D")
    pdf.rect(160, y, 10, 10, round_corners=True, style="D")

    # no fill different corners
    y += 35
    pdf.rect(10, y, 15, 20, round_corners=("TOP_LEFT",), style="D")
    pdf.rect(30, y, 20, 15, round_corners=("TOP_RIGHT",), style="D")
    pdf.rect(60, y, 33, 28, round_corners=("BOTTOM_LEFT",), style="D")
    pdf.rect(100, y, 50, 10, round_corners=("BOTTOM_RIGHT",), style="D")
    pdf.rect(160, y, 10, 10, round_corners=("TOP_LEFT",), style="D")

    # no fill multiple corners
    y += 35
    pdf.rect(10, y, 15, 20, round_corners=("TOP_LEFT", "BOTTOM_RIGHT"), style="D")
    pdf.rect(30, y, 20, 15, round_corners=("TOP_RIGHT", "TOP_LEFT"), style="D")
    pdf.rect(
        60,
        y,
        33,
        28,
        round_corners=("BOTTOM_LEFT", "BOTTOM_RIGHT", "TOP_LEFT"),
        style="D",
    )
    pdf.rect(
        100,
        y,
        50,
        10,
        round_corners=("BOTTOM_RIGHT", "TOP_RIGHT", "TOP_LEFT"),
        style="D",
    )
    pdf.rect(160, y, 10, 10, round_corners=("TOP_LEFT", "BOTTOM_LEFT"), style="D")

    # fill color different sizes test
    pdf.set_fill_color(0, 255, 0)
    y += 35
    pdf.rect(10, y, 15, 20, round_corners=True, style="DF")
    pdf.rect(30, y, 20, 15, round_corners=True, style="DF")
    pdf.rect(60, y, 33, 28, round_corners=True, style="DF")
    pdf.rect(100, y, 50, 10, round_corners=True, style="DF")
    pdf.rect(160, y, 10, 10, round_corners=True, style="DF")

    # fill different corners
    y += 35
    pdf.rect(10, y, 15, 20, round_corners=("TOP_LEFT",), style="DF")
    pdf.rect(30, y, 20, 15, round_corners=("TOP_RIGHT",), style="DF")
    pdf.rect(60, y, 33, 28, round_corners=("BOTTOM_LEFT",), style="DF")
    pdf.rect(100, y, 50, 10, round_corners=("BOTTOM_RIGHT",), style="DF")
    pdf.rect(160, y, 10, 10, round_corners=("TOP_LEFT",), style="DF")

    # fill multiple corners
    y += 35
    pdf.rect(10, y, 15, 20, round_corners=("TOP_LEFT", "BOTTOM_RIGHT"), style="DF")
    pdf.rect(30, y, 20, 15, round_corners=("TOP_RIGHT", "TOP_LEFT"), style="DF")
    pdf.rect(
        60,
        y,
        33,
        28,
        round_corners=("BOTTOM_LEFT", "BOTTOM_RIGHT", "TOP_LEFT"),
        style="DF",
    )
    pdf.rect(
        100,
        y,
        50,
        10,
        round_corners=("BOTTOM_RIGHT", "TOP_RIGHT", "TOP_LEFT"),
        style="DF",
    )
    pdf.rect(160, y, 10, 10, round_corners=("TOP_LEFT", "BOTTOM_LEFT"), style="DF")

    # fill only different corners
    pdf.set_fill_color(255, 255, 0)
    y += 35
    pdf.rect(10, y, 15, 20, round_corners=("TOP_LEFT",), style="F")
    pdf.rect(30, y, 20, 15, round_corners=("TOP_RIGHT",), style="F")
    pdf.rect(60, y, 33, 28, round_corners=("BOTTOM_LEFT",), style="F")
    pdf.rect(100, y, 50, 10, round_corners=("BOTTOM_RIGHT",), style="F")
    pdf.rect(160, y, 10, 10, round_corners=("TOP_LEFT",), style="F")

    # fill only multiple corners
    y += 35
    pdf.rect(10, y, 15, 20, round_corners=("TOP_LEFT", "BOTTOM_RIGHT"), style="F")
    pdf.rect(30, y, 20, 15, round_corners=("TOP_RIGHT", "TOP_LEFT"), style="F")
    pdf.rect(
        60,
        y,
        33,
        28,
        round_corners=("BOTTOM_LEFT", "BOTTOM_RIGHT", "TOP_LEFT"),
        style="F",
        corner_radius=8,
    )
    pdf.rect(
        100,
        y,
        50,
        10,
        round_corners=("BOTTOM_RIGHT", "TOP_RIGHT", "TOP_LEFT"),
        style="F",
    )
    pdf.rect(160, y, 10, 10, round_corners=("TOP_LEFT", "BOTTOM_LEFT"), style="F")

    assert_pdf_equal(pdf, HERE / "class_round_corners_rect.pdf", tmp_path)


def test_round_corners_rect_conditionals(tmp_path):
    pdf = fpdf.FPDF()
    pdf.add_page()

    # for w < h  0 (w = h)
    pdf.rect(10, 10, 20, 40, round_corners=("BOTTOM_LEFT",))

    # for radius == 0 (w = h)
    pdf.rect(40, 10, 30, 30, round_corners=("BOTTOM_LEFT",))

    # for radius >= min / 2 (min < max/2)
    pdf.rect(80, 10, 50, 10, round_corners=True, corner_radius=20)

    assert_pdf_equal(pdf, HERE / "class_round_corners_rect_conditionals.pdf", tmp_path)


def test_round_corners_rect_draw_color(tmp_path):
    pdf = fpdf.FPDF()
    pdf.add_page()

    # changing the draw color
    pdf.set_draw_color(150)
    pdf.rect(10, 10, 60, 80, round_corners=("TOP_LEFT", "BOTTOM_RIGHT"))
    pdf.rect(90, 10, 50, 65, round_corners=("BOTTOM_LEFT",), style="DF")

    pdf.set_draw_color(0, 255, 255)
    pdf.rect(10, 100, 60, 80, round_corners=("TOP_LEFT", "BOTTOM_RIGHT"))
    pdf.rect(90, 100, 50, 65, round_corners=("BOTTOM_LEFT",), style="DF")

    assert_pdf_equal(pdf, HERE / "class_round_corners_rect_draw_color.pdf", tmp_path)


def test_round_corners_rect_invalid_value():
    pdf = fpdf.FPDF()
    pdf.add_page()

    with pytest.raises(ValueError):
        pdf.rect(100, 100, 15, 20, round_corners="BOTTOM_CENTER")



#!/usr/bin/env python3
# USAGE: ./add_on_page_with_pdfrw.py src_file.pdf dest_file.pdf
import sys
from fpdf import FPDF
from pdfrw import PageMerge, PdfReader, PdfWriter
from pdfrw.pagemerge import RectXObj

IN_FILEPATH = sys.argv[1]
OUT_FILEPATH = sys.argv[2]
ON_PAGE_INDEX = 1
# if True, new content will be placed underneath page (painted first):
UNDERNEATH = False

reader = PdfReader(IN_FILEPATH)
area = RectXObj(reader.pages[0])


def new_content():
    fpdf = FPDF(format=(area.w, area.h), unit="pt")
    fpdf.add_page()
    fpdf.set_font("helvetica", size=36)
    fpdf.text(50, 50, "Hello!")
    reader = PdfReader(fdata=bytes(fpdf.output()))
    return reader.pages[0]


writer = PdfWriter()
writer.pagearray = reader.Root.Pages.Kids
if writer.pagearray[0].Kids:
    writer.pagearray = writer.pagearray[0].Kids
PageMerge(writer.pagearray[ON_PAGE_INDEX]).add(
    new_content(), prepend=UNDERNEATH
).render()
writer.write(OUT_FILEPATH)



from pathlib import Path

from fpdf import FPDF, FontFace

from test.conftest import assert_pdf_equal

HERE = Path(__file__).resolve().parent


def test_header_footer_and_use_font_face(tmp_path):  # issue 1204
    class PDF(FPDF):
        def header(self):
            with self.use_font_face(FontFace(color="#00ff00", size_pt=12)):  # LABEL A
                self.cell(text=f"Header {self.page_no()}")
            self.ln()

        def footer(self):
            self.set_y(-15)
            with self.use_font_face(FontFace(color="#0000ff", size_pt=12)):  # LABEL B
                self.cell(text=f"Footer {self.page_no()}")

    pdf = PDF()
    pdf.set_font(family="helvetica", size=12)
    pdf.add_page()
    with pdf.use_font_face(FontFace(size_pt=36)):  # LABEL C
        pdf.multi_cell(w=0, text="\n".join(f"Line {i + 1}" for i in range(21)))
    assert pdf.font_size_pt == 12
    assert_pdf_equal(pdf, HERE / "header_footer_and_use_font_face.pdf", tmp_path)



import pytest

from decimal import Decimal
from contextlib import contextmanager

from fpdf.drawing import (
    Arc,
    BezierCurve,
    Close,
    Ellipse,
    HorizontalLine,
    ImplicitClose,
    Line,
    Move,
    PaintedPath,
    QuadraticBezierCurve,
    Rectangle,
    RelativeArc,
    RelativeBezierCurve,
    RelativeHorizontalLine,
    RelativeLine,
    RelativeMove,
    RelativeQuadraticBezierCurve,
    RelativeVerticalLine,
    RoundedRectangle,
    VerticalLine,
)
from fpdf.drawing_primitives import Point, Transform, rgb8
from fpdf.enums import (
    BlendMode,
    ClippingPathIntersectionRule,
    IntersectionRule,
    PathPaintRule,
    StrokeCapStyle,
    StrokeJoinStyle,
)
from fpdf.syntax import Name, Raw


@contextmanager
def no_exception():
    yield


def exception(exc):
    def wrapper():
        return pytest.raises(exc)

    return wrapper


hex_colors = (
    pytest.param("#CAB", rgb8(r=0xCC, g=0xAA, b=0xBB, a=None), id="RGB hex 3"),
    pytest.param("#FADE", rgb8(r=0xFF, g=0xAA, b=0xDD, a=0xEE), id="RGBA hex 4"),
    pytest.param("#C0FFEE", rgb8(r=0xC0, g=0xFF, b=0xEE, a=None), id="RGB hex 6"),
    pytest.param("#0D06F00D", rgb8(r=0x0D, g=0x06, b=0xF0, a=0x0D), id="RGBA hex 8"),
    pytest.param(
        "#0d06f00d",
        rgb8(r=0x0D, g=0x06, b=0xF0, a=0x0D),
        id="RGBA hex 9, case insensitive",
    ),
    pytest.param("#asd", ValueError, id="bad characters"),
    pytest.param("asd", ValueError, id="bad characters missing hash"),
    pytest.param("#12345", ValueError, id="wrong length"),
    pytest.param("123456", ValueError, id="missing hash"),
    pytest.param(123, TypeError, id="bad type integer"),
)

rgb_colors = (
    pytest.param("rgb(204, 170, 187)", rgb8(r=204, g=170, b=187, a=None), id="RGB"),
    pytest.param("rgb(13, 6, 240, 5)", rgb8(r=13, g=6, b=240, a=5), id="RGBA"),
    pytest.param(
        "rgb( 192,  255 ,  238 )",
        rgb8(r=192, g=255, b=238, a=None),
        id="rgb with extra spaces",
    ),
    pytest.param("rgb(a, s, d)", ValueError, id="bad characters"),
    pytest.param("(240, 170, 187)", ValueError, id="missing rgb"),
    pytest.param("rgb(10, 10, 10", ValueError, id="missing )"),
    pytest.param("rgb(13, 6, 240, 5, 200)", ValueError, id="wrong length"),
    pytest.param("rgba(13, 6, 240, 5)", ValueError, id="rgba instead of rgb"),
    pytest.param(123, TypeError, id="bad type integer"),
)

numbers = (
    pytest.param(100, "100", id="integer"),
    pytest.param(Decimal("1.1"), "1.1", id="Decimal"),
    pytest.param(Decimal("0.000008"), "0", id="truncated Decimal"),
    pytest.param(1.05, "1.05", id="float"),
    pytest.param(10.00001, "10", id="truncated float"),
    pytest.param(-1.12345, "-1.1235", id="rounded float"),
    pytest.param(-0.00004, "-0", id="negative zero"),
)

r = Raw


class CustomPrimitive:
    # pylint: disable=no-self-use
    def serialize(self):
        return "custom primitive"


pdf_primitives = (
    pytest.param(Raw("raw output"), r("raw output"), id="raw"),
    pytest.param(CustomPrimitive(), r("custom primitive"), id="custom"),
    pytest.param(Name("pdf_name"), r("/pdf_name"), id="name"),
    pytest.param(Name("pdf#<name>"), r("/pdf#23#3Cname#3E"), id="escape name"),
    pytest.param("string", r("(string)"), id="string"),
    pytest.param("\r()\\", r(r"""(\r\(\)\\)"""), id="escape string"),
    pytest.param(b"bytes", r("<6279746573>"), id="bytes"),
    pytest.param(123, r("123"), id="integer"),
    pytest.param(123.456, r("123.456"), id="float"),
    pytest.param(Decimal("1.1"), r("1.1"), id="decimal"),
    pytest.param(True, r("true"), id="True"),
    pytest.param(False, r("false"), id="False"),
    pytest.param(None, r("null"), id="None"),
    pytest.param(["a", b"b", 0xC, Name("d")], r("[(a) <62> 12 /d]"), id="list"),
    pytest.param(("a", b"b", 0xC, Name("d")), r("[(a) <62> 12 /d]"), id="tuple"),
    pytest.param(
        {Name("key"): "value", Name("k2"): True},
        r("<< /key (value)\n/k2 true >>"),
        id="dict",
    ),
)

pdf_bad_primitives = (
    pytest.param(type("EmptyClass", (), {}), TypeError, id="Class without bdf_repr"),
    pytest.param({"element", 1, 2.0}, TypeError, id="unsupported (set)"),
    pytest.param({"key": "value", "k2": True}, ValueError, id="dict with bad keys"),
)

T = Transform
P = Point

transforms = (
    pytest.param(T.identity(), P(1, 1), P(1, 1), id="identity"),
    pytest.param(T.translation(x=1, y=1), P(1, 1), P(2, 2), id="translation"),
    pytest.param(T.scaling(x=2, y=3), P(1, 1), P(2, 3), id="scaling x-y"),
    pytest.param(T.scaling(2), P(1, 1), P(2, 2), id="scaling"),
    pytest.param(T.rotation(1.5707963267948966), P(1, 0), P(0, 1.0), id="rotation"),
    pytest.param(T.rotation_d(90), P(1, 0), P(0, 1.0), id="rotation_d"),
    pytest.param(T.shearing(1, 0), P(1, 1), P(2, 1), id="shearing"),
)

coercive_enums = (
    pytest.param(
        IntersectionRule,
        (IntersectionRule.NONZERO, "NONZERO", "nonzero", "NONzero"),
        IntersectionRule.NONZERO,
        no_exception,
        id="IntersectionRule.NONZERO",
    ),
    pytest.param(
        IntersectionRule,
        (IntersectionRule.EVENODD, "EVENODD", "evenodd", "EveNOdD"),
        IntersectionRule.EVENODD,
        no_exception,
        id="IntersectionRule.EVENODD",
    ),
    pytest.param(
        IntersectionRule,
        ("nonsense",),
        None,
        exception(ValueError),
        id="coerce bad string",
    ),
    pytest.param(
        IntersectionRule,
        (1234,),
        None,
        exception(TypeError),
        id="coerce wrong type entirely",
    ),
    pytest.param(
        PathPaintRule,
        (PathPaintRule.STROKE, "stroke", "S"),
        PathPaintRule.STROKE,
        no_exception,
        id="PathPaintRule.STROKE",
    ),
    pytest.param(
        PathPaintRule,
        (PathPaintRule.FILL_NONZERO, "fill_nonzero", "f"),
        PathPaintRule.FILL_NONZERO,
        no_exception,
        id="PathPaintRule.FILL_NONZERO",
    ),
    pytest.param(
        PathPaintRule,
        (PathPaintRule.FILL_EVENODD, "fill_evenodd", "f*"),
        PathPaintRule.FILL_EVENODD,
        no_exception,
        id="PathPaintRule.FILL_EVENODD",
    ),
    pytest.param(
        PathPaintRule,
        (PathPaintRule.STROKE_FILL_NONZERO, "stroke_fill_nonzero", "B"),
        PathPaintRule.STROKE_FILL_NONZERO,
        no_exception,
        id="PathPaintRule.STROKE_FILL_NONZERO",
    ),
    pytest.param(
        PathPaintRule,
        (PathPaintRule.STROKE_FILL_EVENODD, "stroke_fill_evenodd", "B*"),
        PathPaintRule.STROKE_FILL_EVENODD,
        no_exception,
        id="PathPaintRule.STROKE_FILL_EVENODD",
    ),
    pytest.param(
        PathPaintRule,
        (PathPaintRule.DONT_PAINT, "dont_paint", "n"),
        PathPaintRule.DONT_PAINT,
        no_exception,
        id="PathPaintRule.DONT_PAINT",
    ),
    pytest.param(
        PathPaintRule,
        (PathPaintRule.AUTO, "auto"),
        PathPaintRule.AUTO,
        no_exception,
        id="PathPaintRule.AUTO",
    ),
    pytest.param(
        ClippingPathIntersectionRule,
        (ClippingPathIntersectionRule.NONZERO, "nonzero", "W"),
        ClippingPathIntersectionRule.NONZERO,
        no_exception,
        id="ClippingPathIntersectionRule.NONZERO",
    ),
    pytest.param(
        ClippingPathIntersectionRule,
        (ClippingPathIntersectionRule.EVENODD, "evenodd", "W*"),
        ClippingPathIntersectionRule.EVENODD,
        no_exception,
        id="ClippingPathIntersectionRule.EVENODD",
    ),
    pytest.param(
        StrokeCapStyle,
        (-1,),
        None,
        exception(ValueError),
        id="int coerce out of range",
    ),
    pytest.param(
        StrokeCapStyle,
        ("nonsense",),
        None,
        exception(ValueError),
        id="int coerce bad key",
    ),
    pytest.param(
        StrokeCapStyle,
        (1.0, object()),
        None,
        exception(TypeError),
        id="int coerce bad type",
    ),
    pytest.param(
        StrokeCapStyle,
        (StrokeCapStyle.BUTT, "butt", 0),
        StrokeCapStyle.BUTT,
        no_exception,
        id="StrokeCapStyle.BUTT",
    ),
    pytest.param(
        StrokeCapStyle,
        (StrokeCapStyle.ROUND, "round", 1),
        StrokeCapStyle.ROUND,
        no_exception,
        id="StrokeCapStyle.ROUND",
    ),
    pytest.param(
        StrokeCapStyle,
        (StrokeCapStyle.SQUARE, "square", 2),
        StrokeCapStyle.SQUARE,
        no_exception,
        id="StrokeCapStyle.SQUARE",
    ),
    pytest.param(
        StrokeJoinStyle,
        (StrokeJoinStyle.MITER, "miter", 0),
        StrokeJoinStyle.MITER,
        no_exception,
        id="StrokeJoinStyle.MITER",
    ),
    pytest.param(
        StrokeJoinStyle,
        (StrokeJoinStyle.ROUND, "round", 1),
        StrokeJoinStyle.ROUND,
        no_exception,
        id="StrokeJoinStyle.ROUND",
    ),
    pytest.param(
        StrokeJoinStyle,
        (StrokeJoinStyle.BEVEL, "bevel", 2),
        StrokeJoinStyle.BEVEL,
        no_exception,
        id="StrokeJoinStyle.BEVEL",
    ),
    pytest.param(
        BlendMode,
        (BlendMode.NORMAL, "normal"),
        BlendMode.NORMAL,
        no_exception,
        id="BlendMode.NORMAL",
    ),
    pytest.param(
        BlendMode,
        (BlendMode.MULTIPLY, "multiply"),
        BlendMode.MULTIPLY,
        no_exception,
        id="BlendMode.MULTIPLY",
    ),
    pytest.param(
        BlendMode,
        (BlendMode.SCREEN, "screen"),
        BlendMode.SCREEN,
        no_exception,
        id="BlendMode.SCREEN",
    ),
    pytest.param(
        BlendMode,
        (BlendMode.OVERLAY, "overlay"),
        BlendMode.OVERLAY,
        no_exception,
        id="BlendMode.OVERLAY",
    ),
    pytest.param(
        BlendMode,
        (BlendMode.DARKEN, "darken"),
        BlendMode.DARKEN,
        no_exception,
        id="BlendMode.DARKEN",
    ),
    pytest.param(
        BlendMode,
        (BlendMode.LIGHTEN, "lighten"),
        BlendMode.LIGHTEN,
        no_exception,
        id="BlendMode.LIGHTEN",
    ),
    pytest.param(
        BlendMode,
        (BlendMode.COLOR_DODGE, "color_dodge"),
        BlendMode.COLOR_DODGE,
        no_exception,
        id="BlendMode.COLOR_DODGE",
    ),
    pytest.param(
        BlendMode,
        (BlendMode.COLOR_BURN, "color_burn"),
        BlendMode.COLOR_BURN,
        no_exception,
        id="BlendMode.COLOR_BURN",
    ),
    pytest.param(
        BlendMode,
        (BlendMode.HARD_LIGHT, "hard_light"),
        BlendMode.HARD_LIGHT,
        no_exception,
        id="BlendMode.HARD_LIGHT",
    ),
    pytest.param(
        BlendMode,
        (BlendMode.SOFT_LIGHT, "soft_light"),
        BlendMode.SOFT_LIGHT,
        no_exception,
        id="BlendMode.SOFT_LIGHT",
    ),
    pytest.param(
        BlendMode,
        (BlendMode.DIFFERENCE, "difference"),
        BlendMode.DIFFERENCE,
        no_exception,
        id="BlendMode.DIFFERENCE",
    ),
    pytest.param(
        BlendMode,
        (BlendMode.EXCLUSION, "exclusion"),
        BlendMode.EXCLUSION,
        no_exception,
        id="BlendMode.EXCLUSION",
    ),
    pytest.param(
        BlendMode,
        (BlendMode.HUE, "hue"),
        BlendMode.HUE,
        no_exception,
        id="BlendMode.HUE",
    ),
    pytest.param(
        BlendMode,
        (BlendMode.SATURATION, "saturation"),
        BlendMode.SATURATION,
        no_exception,
        id="BlendMode.SATURATION",
    ),
    pytest.param(
        BlendMode,
        (BlendMode.COLOR, "color"),
        BlendMode.COLOR,
        no_exception,
        id="BlendMode.COLOR",
    ),
    pytest.param(
        BlendMode,
        (BlendMode.LUMINOSITY, "luminosity"),
        BlendMode.LUMINOSITY,
        no_exception,
        id="BlendMode.LUMINOSITY",
    ),
)


style_attributes = (
    pytest.param("auto_close", True, id="auto close"),
    pytest.param("intersection_rule", "evenodd", id="intersection rule"),
    pytest.param("fill_color", None, id="no fill color"),
    pytest.param("fill_color", "#00F", id="RGB 3 fill color"),
    pytest.param("fill_color", "#00FC", id="RGBA 4 fill color"),
    pytest.param("fill_color", "#00FF00", id="RGB 6 fill color"),
    pytest.param("fill_color", "#00FF007F", id="RGBA 8 fill color"),
    pytest.param("fill_opacity", None, id="no fill opacity"),
    pytest.param("fill_opacity", 0.5, id="half fill opacity"),
    pytest.param("stroke_color", None, id="no stroke color"),
    pytest.param("stroke_color", "#00F", id="RGB 3 stroke color"),
    pytest.param("stroke_color", "#00FC", id="RGBA 4 stroke color"),
    pytest.param("stroke_color", "#00FF00", id="RGB 6 stroke color"),
    pytest.param("stroke_color", "#00FF007F", id="RGBA 8 stroke color"),
    pytest.param("stroke_opacity", None, id="no stroke opacity"),
    pytest.param("stroke_opacity", 0.5, id="half stroke opacity"),
    pytest.param("stroke_width", None, id="no stroke width"),
    pytest.param("stroke_width", 0, id="0 stroke width"),
    pytest.param("stroke_width", 2, id="2 stroke width"),
    pytest.param("stroke_join_style", "miter", id="miter stroke join"),
    pytest.param("stroke_join_style", "bevel", id="bevel stroke join"),
    pytest.param("stroke_cap_style", "butt", id="butt stroke cap"),
    pytest.param("stroke_cap_style", "square", id="square stroke cap"),
    pytest.param("stroke_dash_pattern", 0.5, id="numeric stroke dash pattern"),
    pytest.param("stroke_dash_pattern", [0.5, 0.5], id="50-50 stroke dash pattern"),
    pytest.param("stroke_dash_pattern", [1, 2, 3, 1], id="complex stroke dash pattern"),
)

blend_modes = (
    pytest.param("normal"),
    pytest.param("multiply"),
    pytest.param("screen"),
    pytest.param("overlay"),
    pytest.param("darken"),
    pytest.param("lighten"),
    pytest.param("color_dodge"),
    pytest.param("color_burn"),
    pytest.param("hard_light"),
    pytest.param("soft_light"),
    pytest.param("difference"),
    pytest.param("exclusion"),
    pytest.param("hue"),
    pytest.param("saturation"),
    pytest.param("color"),
    pytest.param("luminosity"),
)

invalid_styles = (
    pytest.param("auto_close", 2, TypeError, id="invalid auto_close"),
    pytest.param("paint_rule", 123, TypeError, id="invalid numeric paint_rule"),
    pytest.param("paint_rule", "asdasd", ValueError, id="invalid string paint_rule"),
    pytest.param(
        "intersection_rule", 123, TypeError, id="invalid numeric intersection_rule"
    ),
    pytest.param(
        "intersection_rule", "asdasd", ValueError, id="invalid string intersection_rule"
    ),
    pytest.param("fill_color", "123", ValueError, id="invalid string fill_color"),
    pytest.param("fill_color", 2, TypeError, id="invalid numeric fill_color"),
    pytest.param(
        "fill_opacity",
        "123123",
        (TypeError, AssertionError),
        id="invalid string fill_opacity",
    ),
    pytest.param("fill_opacity", 2, ValueError, id="invalid numeric fill_opacity"),
    pytest.param("stroke_color", [2], TypeError, id="invalid stroke_color"),
    pytest.param(
        "stroke_dash_pattern", "123", TypeError, id="invalid string stroke_dash_pattern"
    ),
    pytest.param(
        "stroke_dash_pattern",
        [0.5, "0.5"],
        TypeError,
        id="invalid string in stroke_dash_pattern",
    ),
    pytest.param(
        "stroke_dash_phase", "123", TypeError, id="invalid string stroke_dash_phase"
    ),
    pytest.param(
        "stroke_miter_limit", "123", TypeError, id="invalid string stroke_miter_limit"
    ),
    pytest.param("stroke_width", [2], TypeError, id="invalid stroke_width"),
    pytest.param("invalid_style_name", 2, AttributeError, id="invalid style name"),
)


M = Move
L = Line
B = BezierCurve

path_elements = (
    pytest.param(
        P(1, 1),
        Move(P(1, 2)),
        "1 2 m",
        P(1, 2),
        Move,
        id="move",
    ),
    pytest.param(
        P(1, 1),
        RelativeMove(P(1, 2)),
        "2 3 m",
        P(2, 3),
        Move,
        id="relative move",
    ),
    pytest.param(
        P(1, 1),
        Line(P(1, 2)),
        "1 2 l",
        P(1, 2),
        Line,
        id="line",
    ),
    pytest.param(
        P(1, 1),
        RelativeLine(P(1, 2)),
        "2 3 l",
        P(2, 3),
        Line,
        id="relative line",
    ),
    pytest.param(
        P(1, 1),
        HorizontalLine(2),
        "2 1 l",
        P(2, 1),
        Line,
        id="horizontal line",
    ),
    pytest.param(
        P(1, 1),
        RelativeHorizontalLine(2),
        "3 1 l",
        P(3, 1),
        Line,
        id="relative horizontal line",
    ),
    pytest.param(
        P(1, 1),
        VerticalLine(2),
        "1 2 l",
        P(1, 2),
        Line,
        id="vertical line",
    ),
    pytest.param(
        P(1, 1),
        RelativeVerticalLine(2),
        "1 3 l",
        P(1, 3),
        Line,
        id="relative vertical line",
    ),
    pytest.param(
        P(1, 1),
        BezierCurve(P(1, 2), P(3, 4), P(5, 6)),
        "1 2 3 4 5 6 c",
        P(5, 6),
        BezierCurve,
        id="cubic bezier curve",
    ),
    pytest.param(
        P(1, 1),
        RelativeBezierCurve(P(1, 2), P(3, 4), P(5, 6)),
        "2 3 4 5 6 7 c",
        P(6, 7),
        BezierCurve,
        id="relative cubic bezier curve",
    ),
    pytest.param(
        P(1, 1),
        QuadraticBezierCurve(P(1, 2), P(3, 4)),
        "1 1.6667 1.6667 2.6667 3 4 c",
        P(3, 4),
        QuadraticBezierCurve,
        id="quadratic bezier curve",
    ),
    pytest.param(
        P(1, 1),
        RelativeQuadraticBezierCurve(P(1, 2), P(3, 4)),
        "1.6667 2.3333 2.6667 3.6667 4 5 c",
        P(4, 5),
        QuadraticBezierCurve,
        id="relative quadratic bezier curve",
    ),
    pytest.param(
        P(1, 1),
        Arc(P(1, 2), 0, True, True, P(3, 4)),
        "1.4142 -0.1046 2.1977 -0.3284 2.75 0.5 c 3.3023 1.3284 3.4142 2.8954 3 4 c",
        P(3, 4),
        BezierCurve,
        id="arc",
    ),
    pytest.param(
        P(1, 1),
        RelativeArc(P(1, 2), 0, True, True, P(3, 4)),
        "1.5523 -0.6569 2.6716 -1.1046 3.5 0 c 4.3284 1.1046 4.5523 3.3431 4 5 c",
        P(4, 5),
        BezierCurve,
        id="relative arc",
    ),
    pytest.param(
        P(1, 1),
        Rectangle(P(1, 2), P(3, 4)),
        "1 2 3 4 re",
        P(1, 2),
        Line,
        id="rectangle",
    ),
    pytest.param(
        P(1, 1),
        RoundedRectangle(P(1, 2), P(6, 6), P(1, 2)),
        (
            "2 2 m 6 2 l 6.5523 2 7 2.8954 7 4 c 7 6 l 7 7.1046 6.5523 8 6 8 c 2 8 l "
            "1.4477 8 1 7.1046 1 6 c 1 4 l 1 2.8954 1.4477 2 2 2 c h"
        ),
        P(1, 2),
        Line,
        id="rounded rectangle",
    ),
    pytest.param(
        P(1, 1),
        Ellipse(P(1, 2), P(3, 4)),
        (
            "4 4 m 4 5.1046 3.5523 6 3 6 c 2.4477 6 2 5.1046 2 4 c 2 2.8954 2.4477 2 3 2 c "
            "3.5523 2 4 2.8954 4 4 c h"
        ),
        P(3, 4),
        Move,
        id="rounded rectangle",
    ),
    pytest.param(
        P(1, 1),
        Ellipse(P(1, 2), P(3, 4)),
        (
            "4 4 m 4 5.1046 3.5523 6 3 6 c 2.4477 6 2 5.1046 2 4 c 2 2.8954 2.4477 2 3 2 c "
            "3.5523 2 4 2.8954 4 4 c h"
        ),
        P(3, 4),
        Move,
        id="ellipse",
    ),
    pytest.param(
        P(1, 1),
        ImplicitClose(),
        "",
        P(1, 1),
        Move,
        id="implicit close",
    ),
    pytest.param(
        P(1, 1),
        Close(),
        "h",
        P(1, 1),
        Move,
        id="close",
    ),
)

PP = PaintedPath

paint_rules = (
    pytest.param(PathPaintRule.STROKE, PathPaintRule.STROKE, id="Enum"),
    pytest.param("FILL_EVENODD", PathPaintRule.FILL_EVENODD, id="matching string"),
    pytest.param(
        "stroke_fill_nonzero",
        PathPaintRule.STROKE_FILL_NONZERO,
        id="matching string after uppercasing",
    ),
    pytest.param(None, PathPaintRule.DONT_PAINT, id="None"),
)

clipping_path_result = (
    pytest.param(
        (
            "q 1 1 m 9 1 l 9 9 l 1 9 l h W n 0 0 m 10 0 l 5 10 l h B Q",
            "1 1 m 9 1 l 9 9 l 1 9 l h W n 0 0 m 10 0 l 5 10 l h W n",
        ),
        id="test",
    ),
)

painted_path_elements = (
    # circle, ellipse, rectangle
    pytest.param(
        ((PP.rectangle, (1, 2, 3, 4)),),
        [Move(P(0, 0)), RoundedRectangle(P(1, 2), P(3, 4), P(0, 0))],
        ("q 0 0 m 1 2 3 4 re B Q", "0 0 m 1 2 3 4 re W n"),
        id="rectangle",
    ),
    pytest.param(
        ((PP.circle, (1, 2, 3)),),
        [Move(P(0, 0)), Ellipse(center=P(1, 2), radii=P(3, 3))],
        (
            (
                "q 0 0 m 4 2 m 4 3.6569 2.6569 5 1 5 c -0.6569 5 -2 3.6569 -2 2 c -2 0.3431 "
                "-0.6569 -1 1 -1 c 2.6569 -1 4 0.3431 4 2 c h B Q"
            ),
            (
                "0 0 m 4 2 m 4 3.6569 2.6569 5 1 5 c -0.6569 5 -2 3.6569 -2 2 c -2 0.3431 "
                "-0.6569 -1 1 -1 c 2.6569 -1 4 0.3431 4 2 c h W n"
            ),
        ),
        id="circle",
    ),
    pytest.param(
        ((PP.ellipse, (1, 2, 3, 4)),),
        [Move(P(0, 0)), Ellipse(center=P(1, 2), radii=P(3, 4))],
        (
            (
                "q 0 0 m 4 2 m 4 4.2091 2.6569 6 1 6 c -0.6569 6 -2 4.2091 -2 2 c -2 -0.2091 "
                "-0.6569 -2 1 -2 c 2.6569 -2 4 -0.2091 4 2 c h B Q"
            ),
            (
                "0 0 m 4 2 m 4 4.2091 2.6569 6 1 6 c -0.6569 6 -2 4.2091 -2 2 c -2 -0.2091 "
                "-0.6569 -2 1 -2 c 2.6569 -2 4 -0.2091 4 2 c h W n"
            ),
        ),
        id="ellipse",
    ),
    pytest.param(
        ((PP.line_to, (2, 1)),),
        [Move(P(0, 0)), Line(P(2, 1))],
        ("q 0 0 m 2 1 l h B Q", "0 0 m 2 1 l W n"),
        id="line_to",
    ),
    pytest.param(
        ((PP.line_relative, (2, 1)),),
        [Move(P(0, 0)), RelativeLine(P(2, 1))],
        ("q 0 0 m 2 1 l h B Q", "0 0 m 2 1 l W n"),
        id="line_relative",
    ),
    pytest.param(
        ((PP.horizontal_line_to, (2,)),),
        [Move(P(0, 0)), HorizontalLine(2)],
        ("q 0 0 m 2 0 l h B Q", "0 0 m 2 0 l W n"),
        id="horizontal_line_to",
    ),
    pytest.param(
        ((PP.horizontal_line_relative, (2,)),),
        [Move(P(0, 0)), RelativeHorizontalLine(2)],
        ("q 0 0 m 2 0 l h B Q", "0 0 m 2 0 l W n"),
        id="horizontal_line_relative",
    ),
    pytest.param(
        ((PP.vertical_line_to, (2,)),),
        [Move(P(0, 0)), VerticalLine(2)],
        ("q 0 0 m 0 2 l h B Q", "0 0 m 0 2 l W n"),
        id="vertical_line_to",
    ),
    pytest.param(
        ((PP.vertical_line_relative, (2,)),),
        [Move(P(0, 0)), RelativeVerticalLine(2)],
        ("q 0 0 m 0 2 l h B Q", "0 0 m 0 2 l W n"),
        id="vertical_line_relative",
    ),
    pytest.param(
        ((PP.curve_to, (1, 2, 3, 4, 5, 6)),),
        [Move(P(0, 0)), BezierCurve(P(1, 2), P(3, 4), P(5, 6))],
        ("q 0 0 m 1 2 3 4 5 6 c h B Q", "0 0 m 1 2 3 4 5 6 c W n"),
        id="curve_to",
    ),
    pytest.param(
        ((PP.curve_relative, (1, 2, 3, 4, 5, 6)),),
        [Move(P(0, 0)), RelativeBezierCurve(P(1, 2), P(3, 4), P(5, 6))],
        ("q 0 0 m 1 2 3 4 5 6 c h B Q", "0 0 m 1 2 3 4 5 6 c W n"),
        id="curve_relative",
    ),
    pytest.param(
        ((PP.quadratic_curve_to, (1, 2, 3, 4)),),
        [Move(P(0, 0)), QuadraticBezierCurve(P(1, 2), P(3, 4))],
        (
            "q 0 0 m 0.6667 1.3333 1.6667 2.6667 3 4 c h B Q",
            "0 0 m 0.6667 1.3333 1.6667 2.6667 3 4 c W n",
        ),
        id="quadratic_curve_to",
    ),
    pytest.param(
        ((PP.quadratic_curve_relative, (1, 2, 3, 4)),),
        [Move(P(0, 0)), RelativeQuadraticBezierCurve(P(1, 2), P(3, 4))],
        (
            "q 0 0 m 0.6667 1.3333 1.6667 2.6667 3 4 c h B Q",
            "0 0 m 0.6667 1.3333 1.6667 2.6667 3 4 c W n",
        ),
        id="quadratic_curve_relative",
    ),
    pytest.param(
        ((PP.arc_to, (3, 3, 0, False, False, 1, 1)),),
        [Move(P(0, 0)), Arc(P(3, 3), 0, False, False, P(1, 1))],
        (
            "q 0 0 m 0.2489 0.4083 0.5917 0.7511 1 1 c h B Q",
            "0 0 m 0.2489 0.4083 0.5917 0.7511 1 1 c W n",
        ),
        id="arc_to",
    ),
    pytest.param(
        ((PP.arc_relative, (3, 3, 0, False, False, 1, 1)),),
        [Move(P(0, 0)), RelativeArc(P(3, 3), 0, False, False, P(1, 1))],
        (
            "q 0 0 m 0.2489 0.4083 0.5917 0.7511 1 1 c h B Q",
            "0 0 m 0.2489 0.4083 0.5917 0.7511 1 1 c W n",
        ),
        id="arc_relative",
    ),
    pytest.param(
        ((PP.close, ()),),
        [Move(P(0, 0)), Close()],
        ("q 0 0 m h B Q", "0 0 m h W n"),
        id="close",
    ),
)



#!/usr/bin/env python3
# Context: search_index.json is a file used by lunr.js & mkdocs-material builtin search plugin
# ( https://squidfunk.github.io/mkdocs-material/plugins/search/ )
# in order to provide the search feature in our public documentation website.
# This script adds entries to search_index.json for every <dt>/<dd> entry
# in the HTML files generated by pdoc, in order for them to be indexed by lunr.js
import json
from pathlib import Path

from bs4 import BeautifulSoup

REPO_ROOT = Path(__file__).parent.parent
SEARCH_INDEX_FILEPATH = REPO_ROOT / "public/search/search_index.json"
PDOC_DIRPATH = REPO_ROOT / "public/fpdf/"


def main():
    with SEARCH_INDEX_FILEPATH.open() as search_index_file:
        search_index = json.load(search_index_file)
    docs = search_index["docs"]
    initial_docs_count = len(docs)
    for html_filepath in PDOC_DIRPATH.glob("*.html"):
        docs.extend(extract_docs(html_filepath))
    with SEARCH_INDEX_FILEPATH.open("w") as search_index_file:
        json.dump(search_index, search_index_file)
    print(
        f"{len(docs) - initial_docs_count} entries successfully added to {SEARCH_INDEX_FILEPATH.name} for pdoc API pages"
    )


def extract_docs(html_filepath):
    soup = BeautifulSoup(html_filepath.read_text(), features="lxml")
    page_title = soup.find("title").get_text().strip()
    for dt in soup.find_all("dt"):
        dt_id = dt.get("id")
        if dt_id:
            dt_label = dt.get_text().strip().split("\n")[0]
            dd = dt.next_sibling.next_sibling
            assert dd.name == "dd"
            yield {
                "location": f"fpdf/{html_filepath.name}#{dt_id}",
                "title": f"{page_title} - {dt_label}",
                "text": str(dd),
            }


if __name__ == "__main__":
    main()



"""
HTML renderer

The contents of this module are internal to fpdf2, and not part of the public API.
They may change at any time without prior warning or any deprecation period,
in non-backward-compatible ways.

Usage documentation at: <https://py-pdf.github.io/fpdf2/HTML.html>
"""

# pyright: reportPrivateUsage=false

import logging
import re
import warnings
from html.parser import HTMLParser
from string import ascii_lowercase, ascii_uppercase
from typing import TYPE_CHECKING, Any, Callable, Optional, Union

from .deprecation import get_stack_level
from .drawing_primitives import (
    ColorInput,
    DeviceGray,
    DeviceRGB,
    color_from_hex_string,
    convert_to_device_color,
)
from .enums import Align, CharVPos, TextEmphasis, XPos, YPos
from .errors import FPDFException
from .fonts import FontFace, TextStyle
from .outline import OutlineSection
from .table import Row, Table
from .util import get_scale_factor, int2roman

if TYPE_CHECKING:
    from .fpdf import FPDF
    from .text_region import Paragraph

LOGGER = logging.getLogger(__name__)
MESSAGE_WAITING_WIN1252 = "\x95"  # MESSAGE WAITING character in Windows-1252 encoding
BULLET_UNICODE = "•"  # U+2022
DEGREE_SIGN_WIN1252 = "\xb0"  # DEGREE SIGN character in Windows-1252 encoding
RING_OPERATOR_UNICODE = "∘"  # U+2218
HEADING_TAGS = ("title", "h1", "h2", "h3", "h4", "h5", "h6")
# Some of the margin values below are fractions, in order to be fully backward-compatible,
# and due to the _scale_units() conversion performed in HTML2FPDF constructor below.
# Those constants are formatted as Mixed Fractions, a mathematical representation
# making clear what the closest integer value is.
DEFAULT_TAG_STYLES = {
    # Inline tags are FontFace instances :
    "a": FontFace(color="#00f", emphasis="UNDERLINE"),
    "b": FontFace(emphasis="BOLD"),
    "code": FontFace(family="Courier"),
    "del": FontFace(emphasis="STRIKETHROUGH"),
    "em": FontFace(emphasis="ITALICS"),
    "font": FontFace(),
    "i": FontFace(emphasis="ITALICS"),
    "s": FontFace(emphasis="STRIKETHROUGH"),
    "strong": FontFace(emphasis="BOLD"),
    "u": FontFace(emphasis="UNDERLINE"),
    # Block tags are TextStyle instances :
    "blockquote": TextStyle(color="#64002d", t_margin=3, b_margin=3),
    "center": TextStyle(t_margin=4 + 7 / 30, l_margin=Align.C),
    "dd": TextStyle(l_margin=10),
    "dt": TextStyle(font_style="B", t_margin=4 + 7 / 30),
    "title": TextStyle(  # Only rendered if render_title_tag=True
        b_margin=0.4,
        font_size_pt=30,
        t_margin=6,
        l_margin="Center",
    ),
    "h1": TextStyle(
        color="#960000", b_margin=0.4, font_size_pt=24, t_margin=5 + 834 / 900
    ),
    "h2": TextStyle(
        color="#960000", b_margin=0.4, font_size_pt=18, t_margin=5 + 453 / 900
    ),
    "h3": TextStyle(
        color="#960000", b_margin=0.4, font_size_pt=14, t_margin=5 + 199 / 900
    ),
    "h4": TextStyle(
        color="#960000", b_margin=0.4, font_size_pt=12, t_margin=5 + 72 / 900
    ),
    "h5": TextStyle(
        color="#960000", b_margin=0.4, font_size_pt=10, t_margin=5 - 55 / 900
    ),
    "h6": TextStyle(
        color="#960000", b_margin=0.4, font_size_pt=8, t_margin=5 - 182 / 900
    ),
    "li": TextStyle(l_margin=5, t_margin=2),
    "p": TextStyle(),
    "pre": TextStyle(t_margin=4 + 7 / 30, font_family="Courier"),
    "ol": TextStyle(t_margin=2),
    "ul": TextStyle(t_margin=2),
}
INLINE_TAGS = ("a", "b", "code", "del", "em", "font", "i", "s", "strong", "u")
BLOCK_TAGS = HEADING_TAGS + (
    "blockquote",
    "center",
    "dd",
    "dt",
    "li",
    "p",
    "pre",
    "ol",
    "ul",
)
# This defensive programming check ensures that we do not forget any tag in the 2 *_TAGS constants above:
assert (set(BLOCK_TAGS) ^ set(INLINE_TAGS)) == set(DEFAULT_TAG_STYLES.keys())

# Pattern to substitute whitespace sequences with a single space character each.
# The following are all Unicode characters with White_Space classification plus the newline.
# The pattern excludes the non-breaking spaces that are included in "\s".
# We also exclude the OGHAM SPACE MARK for now, because while being a word separator,
# it is usually a graphically visible glyph.
_WS_CHARS = "".join(
    (
        # "\u0009",  # CHARACTER TABULATION
        # "\u000a",  # LINE FEED
        # "\u000b",  # LINE TABULATION
        # "\u000c",  # FORM FEED
        # "\u000d",  # CARRIAGE RETURN
        "\u0009-\u000d",  # combine the above
        "\u0020",  # SPACE
        "\u0085",  # NEXT LINE
        # "\u00a0",  # NO-BREAK SPACE   (keep)
        # "\u1680",  # OGHAM SPACE MARK (not actually white)
        # "\u2000",  # EN QUAD
        # "\u2001",  # EM QUAD
        # "\u2002",  # EN SPACE
        # "\u2003",  # EM SPACE
        # "\u2004",  # THREE-PER-EM SPACE
        # "\u2005",  # FOUR-PER-EM SPACE
        # "\u2006",  # SIX-PER-EM SPACE
        # "\u2007",  # FIGURE SPACE
        # "\u2008",  # PUNCTUATION SPACE
        # "\u2009",  # THIN SPACE
        # "\u200a",  # HAIR SPACE
        "\u2000-\u200a",  # combine the above
        "\u2028",  # LINE SEPARATOR
        "\u2029",  # PARAGRAPH SEPARATOR
        # "\u202f",  # NARROW NO-BREAK SPACE (keep)
        "\u205f",  # MEDIUM MATHEMATICAL SPACE
        "\u3000",  # IDEOGRAPHIC SPACE
    )
)
_WS_SUB_PAT = re.compile(f"[{_WS_CHARS}]+")

COLOR_DICT = {
    "black": "#000000",
    "navy": "#000080",
    "darkblue": "#00008b",
    "mediumblue": "#0000cd",
    "blue": "#0000ff",
    "darkgreen": "#006400",
    "green": "#008000",
    "teal": "#008080",
    "darkcyan": "#008b8b",
    "deepskyblue": "#00bfff",
    "darkturquoise": "#00ced1",
    "mediumspringgreen": "#00fa9a",
    "lime": "#00ff00",
    "springgreen": "#00ff7f",
    "aqua": "#00ffff",
    "cyan": "#00ffff",
    "midnightblue": "#191970",
    "dodgerblue": "#1e90ff",
    "lightseagreen": "#20b2aa",
    "forestgreen": "#228b22",
    "seagreen": "#2e8b57",
    "darkslategray": "#2f4f4f",
    "darkslategrey": "#2f4f4f",
    "limegreen": "#32cd32",
    "mediumseagreen": "#3cb371",
    "turquoise": "#40e0d0",
    "royalblue": "#4169e1",
    "steelblue": "#4682b4",
    "darkslateblue": "#483d8b",
    "mediumturquoise": "#48d1cc",
    "indigo": "#4b0082",
    "darkolivegreen": "#556b2f",
    "cadetblue": "#5f9ea0",
    "cornflowerblue": "#6495ed",
    "rebeccapurple": "#663399",
    "mediumaquamarine": "#66cdaa",
    "dimgray": "#696969",
    "dimgrey": "#696969",
    "slateblue": "#6a5acd",
    "olivedrab": "#6b8e23",
    "slategray": "#708090",
    "slategrey": "#708090",
    "lightslategray": "#778899",
    "lightslategrey": "#778899",
    "mediumslateblue": "#7b68ee",
    "lawngreen": "#7cfc00",
    "chartreuse": "#7fff00",
    "aquamarine": "#7fffd4",
    "maroon": "#800000",
    "purple": "#800080",
    "olive": "#808000",
    "gray": "#808080",
    "grey": "#808080",
    "skyblue": "#87ceeb",
    "lightskyblue": "#87cefa",
    "blueviolet": "#8a2be2",
    "darkred": "#8b0000",
    "darkmagenta": "#8b008b",
    "saddlebrown": "#8b4513",
    "darkseagreen": "#8fbc8f",
    "lightgreen": "#90ee90",
    "mediumpurple": "#9370db",
    "darkviolet": "#9400d3",
    "palegreen": "#98fb98",
    "darkorchid": "#9932cc",
    "yellowgreen": "#9acd32",
    "sienna": "#a0522d",
    "brown": "#a52a2a",
    "darkgray": "#a9a9a9",
    "darkgrey": "#a9a9a9",
    "lightblue": "#add8e6",
    "greenyellow": "#adff2f",
    "paleturquoise": "#afeeee",
    "lightsteelblue": "#b0c4de",
    "powderblue": "#b0e0e6",
    "firebrick": "#b22222",
    "darkgoldenrod": "#b8860b",
    "mediumorchid": "#ba55d3",
    "rosybrown": "#bc8f8f",
    "darkkhaki": "#bdb76b",
    "silver": "#c0c0c0",
    "mediumvioletred": "#c71585",
    "indianred": "#cd5c5c",
    "peru": "#cd853f",
    "chocolate": "#d2691e",
    "tan": "#d2b48c",
    "lightgray": "#d3d3d3",
    "lightgrey": "#d3d3d3",
    "thistle": "#d8bfd8",
    "orchid": "#da70d6",
    "goldenrod": "#daa520",
    "palevioletred": "#db7093",
    "crimson": "#dc143c",
    "gainsboro": "#dcdcdc",
    "plum": "#dda0dd",
    "burlywood": "#deb887",
    "lightcyan": "#e0ffff",
    "lavender": "#e6e6fa",
    "darksalmon": "#e9967a",
    "violet": "#ee82ee",
    "palegoldenrod": "#eee8aa",
    "lightcoral": "#f08080",
    "khaki": "#f0e68c",
    "aliceblue": "#f0f8ff",
    "honeydew": "#f0fff0",
    "azure": "#f0ffff",
    "sandybrown": "#f4a460",
    "wheat": "#f5deb3",
    "beige": "#f5f5dc",
    "whitesmoke": "#f5f5f5",
    "mintcream": "#f5fffa",
    "ghostwhite": "#f8f8ff",
    "salmon": "#fa8072",
    "antiquewhite": "#faebd7",
    "linen": "#faf0e6",
    "lightgoldenrodyellow": "#fafad2",
    "oldlace": "#fdf5e6",
    "red": "#ff0000",
    "fuchsia": "#ff00ff",
    "magenta": "#ff00ff",
    "deeppink": "#ff1493",
    "orangered": "#ff4500",
    "tomato": "#ff6347",
    "hotpink": "#ff69b4",
    "coral": "#ff7f50",
    "darkorange": "#ff8c00",
    "lightsalmon": "#ffa07a",
    "orange": "#ffa500",
    "lightpink": "#ffb6c1",
    "pink": "#ffc0cb",
    "gold": "#ffd700",
    "peachpuff": "#ffdab9",
    "navajowhite": "#ffdead",
    "moccasin": "#ffe4b5",
    "bisque": "#ffe4c4",
    "mistyrose": "#ffe4e1",
    "blanchedalmond": "#ffebcd",
    "papayawhip": "#ffefd5",
    "lavenderblush": "#fff0f5",
    "seashell": "#fff5ee",
    "cornsilk": "#fff8dc",
    "lemonchiffon": "#fffacd",
    "floralwhite": "#fffaf0",
    "snow": "#fffafa",
    "yellow": "#ffff00",
    "lightyellow": "#ffffe0",
    "ivory": "#fffff0",
    "white": "#ffffff",
}


def color_as_decimal(
    color: Optional[str] = "#000000",
) -> Optional[DeviceRGB | DeviceGray]:
    """
    Convert a web color name to a (R, G, B) color tuple.
    cf. https://en.wikipedia.org/wiki/Web_colors#HTML_color_names
    """
    if not color:
        return None
    # Checks if color is a name and gets the hex value
    hexcolor = COLOR_DICT.get(color.lower(), color)
    return color_from_hex_string(hexcolor)


def parse_css_style(style_attr: str) -> dict[str, str]:
    """Parse `style="..."` HTML attributes, and return a dict of key-value"""
    style: dict[str, str] = {}
    for element in style_attr.split(";"):
        if not element:
            continue
        pair = element.split(":")
        if len(pair) == 2 and pair[0] and pair[1]:
            attr, value = pair
            style[attr.strip()] = value.strip()
    return style


class HTML2FPDF(HTMLParser):
    "Render basic HTML to FPDF"

    HTML_UNCLOSED_TAGS = ("br", "dd", "dt", "hr", "img", "li", "td", "tr")
    TABLE_LINE_HEIGHT = 1.3

    def __init__(
        self,
        pdf: "FPDF",
        image_map: Callable[[str], str] = lambda src: src,
        li_tag_indent: Optional[float] = None,
        dd_tag_indent: Optional[float] = None,
        table_line_separators: bool = False,
        ul_bullet_char: str = "disc",
        li_prefix_color: ColorInput = (190, 0, 0),
        heading_sizes: Optional[dict[str, float]] = None,
        pre_code_font: Optional[str] = None,
        warn_on_tags_not_matching: bool = True,
        tag_indents: Optional[dict[str, float]] = None,
        tag_styles: Optional[dict[str, TextStyle]] = None,
        font_family: str = "times",
        render_title_tag: bool = False,
    ) -> None:
        """
        Args:
            pdf (fpdf.fpdf.FPDF): an instance of `FPDF`
            image_map (function): an optional one-argument function that map `<img>` "src" to new image URLs
            li_tag_indent (int): [**DEPRECATED since v2.7.9**]
                numeric indentation of `<li>` elements - Set `tag_styles` instead
            dd_tag_indent (int): [**DEPRECATED since v2.7.9**]
                numeric indentation of `<dd>` elements - Set `tag_styles` instead
            table_line_separators (bool): enable horizontal line separators in `<table>`. Defaults to `False`.
            ul_bullet_char (str): bullet character preceding `<li>` items in `<ul>` lists.
                You can also specify special bullet names like `"circle"` or `"disc"` (the default).
                Can also be configured using the HTML `type` attribute of `<ul>` tags.
            li_prefix_color (tuple, str, fpdf.drawing.DeviceCMYK, fpdf.drawing.DeviceGray, fpdf.drawing.DeviceRGB): color for bullets
                or numbers preceding `<li>` tags. This applies to both `<ul>` & `<ol>` lists.
            heading_sizes (dict): [**DEPRECATED since v2.7.9**]
                font size per heading level names ("h1", "h2"...) - Set `tag_styles` instead
            pre_code_font (str): [**DEPRECATED since v2.7.9**]
                font to use for `<pre>` & `<code>` blocks - Set `tag_styles` instead
            warn_on_tags_not_matching (bool): control warnings production for unmatched HTML tags. Defaults to `True`.
            tag_indents (dict): [**DEPRECATED since v2.8.0**]
                mapping of HTML tag names to numeric values representing their horizontal left indentation. - Set `tag_styles` instead
            tag_styles (dict[str, fpdf.fonts.TextStyle]): mapping of HTML tag names to `fpdf.TextStyle` or `fpdf.FontFace` instances
            font_family (str): optional font family. Default to Times.
            render_title_tag (bool): Render the document <title> at the beginning of the PDF. Default to False.
        """
        super().__init__()
        self.pdf = pdf
        self.ul_bullet_char = ul_bullet_char
        self.image_map = image_map
        self.li_prefix_color = (
            color_as_decimal(li_prefix_color)
            if isinstance(li_prefix_color, str)
            else convert_to_device_color(li_prefix_color)
        )
        self.warn_on_tags_not_matching = warn_on_tags_not_matching
        # The following 4 attributes are there to serve as "temporary state",
        # so that changes to those settings are saved,
        # but not reflected onto self.pdf yet,
        # and only "effectively" applied when self._write_paragraph() is called.
        # This way, we often avoid useless operators in the PDF content stream.
        self.font_family: str = pdf.font_family or font_family
        self.font_size_pt: float = pdf.font_size_pt
        self.font_emphasis = TextEmphasis.NONE
        self.font_color = pdf.text_color
        # For historical / backward-compatibility reasons,
        # write_html() sets an active font (Times by default):
        self.pdf.set_font(
            family=self.font_family,
            size=self.font_size_pt,
            style=self.font_emphasis.style,
        )
        self.style_stack: list[FontFace] = []  # list of FontFace
        self._page_break_after_paragraph = False
        self.follows_trailing_space = False  # The last write has ended with a space.
        self.follows_heading = False  # We don't want extra space below a heading.
        self.align: Optional[Union[float, Align]] = None
        self.heading_level: Optional[int] = None
        self._tags_stack: list[str] = []
        self._column = self.pdf.text_columns(skip_leading_spaces=True)
        self._paragraph: Optional["Paragraph"] = self._column.paragraph()
        # <pre>-related properties:
        self._pre_formatted = False  # preserve whitespace while True.
        # nothing written yet to <pre>, remove one initial nl:
        self._pre_started = False
        # <a>-related properties:
        self.href: int | str = ""
        # <ul>/<ol>-related properties:
        self.indent = 0
        self.line_height_stack: list[float | None] = []
        self.ol_type: dict[int, str] = (
            {}
        )  # when inside a <ol> tag, can be "a", "A", "i", "I" or "1"
        self.bullet: list[int | str] = []
        # <title>-related properties:
        self.render_title_tag = render_title_tag
        self._in_title = False
        # <table>-related properties:
        self.table_line_separators = table_line_separators
        self.table: Optional[Table] = (
            None  # becomes a Table instance when processing <table> tags
        )
        self.table_row: Optional[Row] = (
            None  # becomes a Row instance when processing <tr> tags
        )
        self.tr: Optional[dict[str, Optional[str]]] = (
            None  # becomes a dict of attributes when processing <tr> tags
        )
        self.td_th: Optional[dict[str, Any]] = (
            None  # becomes a dict of attributes when processing <td>/<th> tags
        )
        #                    "inserted" is a special attribute indicating that a cell has be inserted in self.table_row

        self.tag_styles = _scale_units(pdf, DEFAULT_TAG_STYLES)
        for tag, tag_style in (tag_styles or {}).items():
            if tag not in DEFAULT_TAG_STYLES:
                raise NotImplementedError(
                    f"Cannot set style for HTML tag <{tag}> (contributions are welcome to add support for this)"
                )
            if not isinstance(tag_style, FontFace):
                raise ValueError(
                    f"tag_styles values must be instances of FontFace or TextStyle - received: {tag_style}"
                )
            # We convert FontFace values provided for block tags into TextStyle values:
            if tag in BLOCK_TAGS and not isinstance(tag_style, TextStyle):
                # pylint: disable=redefined-loop-name
                tag_style = TextStyle(
                    font_family=tag_style.family,
                    font_style=(
                        "" if not tag_style.emphasis else tag_style.emphasis.style
                    ),
                    font_size_pt=tag_style.size_pt,
                    color=tag_style.color,
                    fill_color=tag_style.fill_color,
                    # Using default tag margins:
                    t_margin=self.tag_styles[
                        tag
                    ].t_margin,  # pyright: ignore[reportAttributeAccessIssue, reportUnknownMemberType, reportUnknownArgumentType]
                    l_margin=self.tag_styles[
                        tag
                    ].l_margin,  # pyright: ignore[reportAttributeAccessIssue, reportUnknownMemberType, reportUnknownArgumentType]
                    b_margin=self.tag_styles[
                        tag
                    ].b_margin,  # pyright: ignore[reportAttributeAccessIssue, reportUnknownMemberType, reportUnknownArgumentType]
                )
            self.tag_styles[tag] = tag_style
        if heading_sizes is not None:
            warnings.warn(
                (
                    "The heading_sizes parameter is deprecated since v2.7.9 "
                    "and will be removed in a future release. "
                    "Set the `tag_styles` parameter instead."
                ),
                DeprecationWarning,
                stacklevel=get_stack_level(),
            )
            for tag, size in heading_sizes.items():
                self.tag_styles[tag] = self.tag_styles[tag].replace(font_size_pt=size)
        if pre_code_font is not None:
            warnings.warn(
                (
                    "The pre_code_font parameter is deprecated since v2.7.9 "
                    "and will be removed in a future release. "
                    "Set the `tag_styles` parameter instead."
                ),
                DeprecationWarning,
                stacklevel=get_stack_level(),
            )
            code_style = self.tag_styles["code"]
            pre_style = self.tag_styles["pre"]
            if isinstance(code_style, TextStyle):
                code_style = code_style.replace(font_family=pre_code_font)
            else:
                code_style = code_style.replace(family=pre_code_font)
            if isinstance(pre_style, TextStyle):
                pre_style = pre_style.replace(font_family=pre_code_font)
            else:
                pre_style = pre_style.replace(family=pre_code_font)
            self.tag_styles["code"] = code_style
            self.tag_styles["pre"] = pre_style
        if dd_tag_indent is not None:
            warnings.warn(
                (
                    "The dd_tag_indent parameter is deprecated since v2.7.9 "
                    "and will be removed in a future release. "
                    "Set the `tag_styles` parameter instead."
                ),
                DeprecationWarning,
                stacklevel=get_stack_level(),
            )
            self.tag_styles["dd"] = self.tag_styles["dd"].replace(
                l_margin=dd_tag_indent
            )
        if li_tag_indent is not None:
            warnings.warn(
                (
                    "The li_tag_indent parameter is deprecated since v2.7.9 "
                    "and will be removed in a future release. "
                    "Set the `tag_styles` parameter instead."
                ),
                DeprecationWarning,
                stacklevel=get_stack_level(),
            )
            self.tag_styles["li"] = self.tag_styles["li"].replace(
                l_margin=li_tag_indent
            )
        if tag_indents:
            warnings.warn(
                (
                    "The tag_indents parameter is deprecated since v2.8.0 "
                    "and will be removed in a future release. "
                    "Set the `tag_styles` parameter instead."
                ),
                DeprecationWarning,
                stacklevel=get_stack_level(),
            )
            for tag, indent in tag_indents.items():
                if tag not in self.tag_styles:
                    raise NotImplementedError(
                        f"Cannot set style for HTML tag <{tag}> (contributions are welcome to add support for this)"
                    )
                self.tag_styles[tag] = self.tag_styles[tag].replace(l_margin=indent)

    @staticmethod
    def _normalize_l_margin(l_margin: Any) -> Union[float, Align]:
        if isinstance(l_margin, Align):
            return l_margin
        if isinstance(l_margin, (int, float)):
            return float(l_margin)
        return 0.0

    def _new_paragraph(
        self,
        align: Optional[Union[float, Align]] = None,
        line_height: Optional[float] = 1.0,
        top_margin: float = 0,
        bottom_margin: float = 0,
        indent: Union[float, Align] = 0,
        bullet: str = "",
    ) -> None:
        # Note that currently top_margin is ignored if bullet is also provided,
        # due to the behaviour of TextRegion._render_column_lines()
        self._end_paragraph()
        self.align = align
        if isinstance(indent, Align):
            # Explicit alignment takes priority over alignment provided as TextStyle.l_margin:
            if not self.align:
                self.align = indent
            indent = 0
        if not top_margin and not self.follows_heading:
            top_margin = self.font_size_pt / self.pdf.k
        self._paragraph = self._column.paragraph(
            text_align=self.align if isinstance(self.align, Align) else None,
            line_height=line_height,
            skip_leading_spaces=True,
            top_margin=top_margin,
            bottom_margin=bottom_margin,
            indent=indent,
            bullet_string=bullet,
        )
        self.follows_trailing_space = True
        self.follows_heading = False

    def _end_paragraph(self) -> None:
        self.align = None
        if not self._paragraph:
            return
        self._column.end_paragraph()
        self._column.render()
        self._paragraph = None
        self.follows_trailing_space = True
        if self._page_break_after_paragraph:
            # pylint: disable=protected-access
            self.pdf._perform_page_break()  # pyright: ignore[reportPrivateUsage]
            self._page_break_after_paragraph = False

    def _write_paragraph(self, text: str, link: Optional[int | str] = None) -> None:
        if not text:
            return
        if not self._paragraph:
            self._new_paragraph()
        # The following local stack is required
        # in order for FPDF._get_current_graphics_state()
        # to properly capture the current graphics state,
        # and then to be able to drop those temporary changes,
        # because they will only be "effectively" applied in .end_paragraph().
        # pylint: disable=protected-access
        self.pdf._push_local_stack()  # pyright: ignore[reportPrivateUsage]
        prev_page = self.pdf.page
        self.pdf.page = 0
        self.pdf.set_font(
            family=self.font_family,
            size=self.font_size_pt,
            style=self.font_emphasis.style,
        )
        if self.font_color != self.pdf.text_color and self.font_color is not None:
            self.pdf.set_text_color(self.font_color)
        assert self._paragraph is not None
        self._paragraph.write(text, link=link)
        self.pdf.page = prev_page
        self.pdf._pop_local_stack()  # pyright: ignore[reportPrivateUsage]

    def _ln(self, h: Optional[float] = None) -> None:
        if self._paragraph:
            self._paragraph.ln(h=h)
        else:
            self._column.ln(h=h)
        self.follows_trailing_space = True

    def handle_data(self, data: str) -> None:
        if self._in_title:
            if self.pdf.title:
                LOGGER.warning('Ignoring repeated <title> "%s"', data)
            else:
                self.pdf.set_title(data)
            if not self.render_title_tag:
                return
        if self.td_th is not None:
            data = data.strip()
            if not data:
                return
            if "inserted" in self.td_th:
                td_th_tag = self.td_th["tag"]
                raise NotImplementedError(
                    f"Unsupported nested HTML tags inside <{td_th_tag}> element: <{self._tags_stack[-1]}>"
                )
                # We could potentially support nested <b> / <em> / <font> tags
                # by building a list of Fragment instances from the HTML cell content
                # and then passing those fragments to Row.cell().
                # However there should be an incoming refactoring of this code
                # dedicated to text layout, and we should probably wait for that
                # before supporting this feature.
            align = self.td_th.get("align")
            if align is None and self.tr is not None:
                align = self.tr.get("align")
            if align:
                align = align.upper()
            bgcolor_str = self.td_th.get("bgcolor")
            if bgcolor_str is None and self.tr is not None:
                bgcolor_str = self.tr.get("bgcolor")
            bgcolor = color_as_decimal(bgcolor_str)
            colspan = int(self.td_th.get("colspan") or "1")
            rowspan = int(self.td_th.get("rowspan") or "1")
            emphasis = 0
            if self.td_th.get("b"):
                emphasis |= TextEmphasis.B
            if self.td_th.get("i"):
                emphasis |= TextEmphasis.I
            if self.td_th.get("U"):
                emphasis |= TextEmphasis.U
            font_family = (
                self.font_family if self.font_family != self.pdf.font_family else None
            )
            font_size_pt = (
                self.font_size_pt
                if self.font_size_pt != self.pdf.font_size_pt
                else None
            )
            font_style = None
            if font_family or emphasis or font_size_pt or bgcolor:
                font_style = FontFace(
                    family=font_family,
                    emphasis=emphasis,
                    size_pt=font_size_pt,
                    color=self.pdf.text_color,
                    fill_color=bgcolor,
                )
            assert self.table_row is not None
            self.table_row.cell(
                text=data,
                align=align,
                style=font_style,
                colspan=colspan,
                rowspan=rowspan,
            )
            self.td_th["inserted"] = True
        elif self.table is not None:
            # ignore anything else than td inside a table
            pass
        elif self._pre_formatted:  # pre blocks
            # If we want to mimic the exact HTML semantics about newlines at the
            # beginning and end of the block, then this needs some more thought.
            if data.startswith("\n") and self._pre_started:
                if data.endswith("\n"):
                    data = data[1:-1]
                else:
                    data = data[1:]
            self._pre_started = False
            self._write_data(data)
        else:
            data = _WS_SUB_PAT.sub(" ", data)
            if self.follows_trailing_space and data[0] == " ":
                self._write_data(data[1:])
            else:
                self._write_data(data)
            self.follows_trailing_space = data[-1] == " "
        if self._page_break_after_paragraph:
            self._end_paragraph()

    def _write_data(self, data: str) -> None:
        if self.href:
            self.put_link(data)
        else:
            if self.heading_level:
                if self.pdf.section_title_styles:
                    raise NotImplementedError(
                        "Combining write_html() & section styles is currently not supported."
                        " You can open up an issue on github.com/py-pdf/fpdf2 if this is something you would like to see implemented."
                    )
                self.pdf.start_section(data, self.heading_level - 1, strict=False)
            self._write_paragraph(data)

    def handle_starttag(self, tag: str, attrs: list[tuple[str, str | None]]) -> None:
        self._pre_started = False
        attrs_dict: dict[str, str | None] = dict(attrs)
        del attrs
        css_style = parse_css_style(attrs_dict.get("style") or "")
        self._tags_stack.append(tag)
        if css_style.get("break-before") == "page":
            self._end_paragraph()
            # pylint: disable=protected-access
            self.pdf._perform_page_break()  # pyright: ignore[reportPrivateUsage]
        if tag in ("b", "i", "u") and self.td_th is not None:
            self.td_th[tag] = True
        if tag == "a":
            self.href = attrs_dict["href"] or ""
            try:
                page = int(self.href)
                self.href = self.pdf.add_link(page=page)
            except ValueError:
                pass
        if tag == "br":
            self._write_paragraph("\n")
        if tag == "hr":
            self._end_paragraph()
            width_str = css_style.get("width", attrs_dict.get("width"))
            if width_str:
                if width_str[-1] == "%":
                    hr_width = self.pdf.epw * float(width_str[:-1]) / 100
                else:
                    hr_width = float(width_str) / self.pdf.k
            else:
                hr_width = self.pdf.epw
            # Centering:
            x_start = self.pdf.l_margin + (self.pdf.epw - hr_width) / 2
            self.pdf.line(
                x1=x_start,
                y1=self.pdf.y,
                x2=x_start + hr_width,
                y2=self.pdf.y,
            )
            self._write_paragraph("\n")
        if tag == "p":
            self.style_stack.append(
                FontFace(
                    family=self.font_family,
                    emphasis=self.font_emphasis,
                    size_pt=self.font_size_pt,
                    color=self.font_color,
                )
            )
            align: Optional[Align] = None
            if "align" in attrs_dict:
                try:
                    align_str = attrs_dict.get("align")
                    align = Align.coerce(align_str) if align_str is not None else None
                except ValueError:
                    align = None
            line_height_str = css_style.get(
                "line-height", attrs_dict.get("line-height")
            )
            # "line-height" attributes are not valid in HTML,
            # but we support it for backward compatibility,
            # because fpdf2 honors it since 2.6.1 and PR #629
            line_height: Optional[float] = None
            if line_height_str:
                try:
                    # YYY parse and convert non-float line_height values
                    line_height = float(line_height_str)
                except ValueError:
                    line_height = None
            tag_style = self.tag_styles[tag]
            if tag_style.color is not None and tag_style.color != self.font_color:
                self.font_color = tag_style.color
            if tag_style.family is not None and tag_style.family != self.font_family:
                self.font_family = tag_style.family
            if tag_style.size_pt is not None and tag_style.size_pt != self.font_size_pt:
                self.font_size_pt = tag_style.size_pt
            if tag_style.emphasis:
                self.font_emphasis |= tag_style.emphasis
            if isinstance(tag_style, TextStyle):
                t_margin = tag_style.t_margin
                b_margin = tag_style.b_margin
                l_margin = tag_style.l_margin
            else:
                t_margin = b_margin = l_margin = 0.0
            l_margin = self._normalize_l_margin(l_margin)
            self._new_paragraph(
                align=align,
                line_height=line_height or 1.0,
                top_margin=t_margin,
                bottom_margin=b_margin,
                indent=l_margin,
            )
        if tag in HEADING_TAGS:
            self.style_stack.append(
                FontFace(
                    family=self.font_family,
                    emphasis=self.font_emphasis,
                    size_pt=self.font_size_pt,
                    color=self.font_color,
                )
            )
            self.heading_level = 0 if tag == "title" else int(tag[1:])
            tag_style = self.tag_styles[tag]
            hsize = (tag_style.size_pt or self.font_size_pt) / self.pdf.k
            align = None
            if "align" in attrs_dict:
                try:
                    align_str = attrs_dict.get("align")
                    align = Align.coerce(align_str) if align_str is not None else None
                except ValueError:
                    align = None
            if isinstance(tag_style, TextStyle):
                t_margin = tag_style.t_margin
                b_margin = tag_style.b_margin
                l_margin = tag_style.l_margin
            else:
                t_margin = b_margin = l_margin = 0.0
            l_margin = self._normalize_l_margin(l_margin)
            self._new_paragraph(
                align=align,
                top_margin=t_margin,
                bottom_margin=b_margin * hsize,
                indent=l_margin,
            )
            if "color" in css_style:
                self.font_color = color_as_decimal(css_style["color"])
            elif "color" in attrs_dict:
                # "color" attributes are not valid in HTML,
                # but we support it for backward compatibility:
                self.font_color = color_as_decimal(attrs_dict["color"])
            elif tag_style.color is not None and tag_style.color != self.font_color:
                self.font_color = tag_style.color
            if tag_style.family is not None and tag_style.family != self.font_family:
                self.font_family = tag_style.family
            if tag_style.size_pt is not None and tag_style.size_pt != self.font_size_pt:
                self.font_size_pt = tag_style.size_pt
            if tag_style.emphasis:
                self.font_emphasis |= tag_style.emphasis
        if tag in (
            "b",
            "blockquote",
            "center",
            "code",
            "del",
            "em",
            "i",
            "dd",
            "dt",
            "pre",
            "s",
            "strong",
            "u",
        ):
            if tag in BLOCK_TAGS:
                self._end_paragraph()
            self.style_stack.append(
                FontFace(
                    family=self.font_family,
                    emphasis=self.font_emphasis,
                    size_pt=self.font_size_pt,
                    color=self.font_color,
                )
            )
            tag_style = self.tag_styles[tag]
            if tag_style.color:
                self.font_color = tag_style.color
            self.font_family = tag_style.family or self.font_family
            self.font_size_pt = tag_style.size_pt or self.font_size_pt
            if tag_style.emphasis:
                self.font_emphasis |= tag_style.emphasis
            if tag == "pre":
                self._pre_formatted = True
                self._pre_started = True
            if tag in BLOCK_TAGS:
                if tag == "dd":
                    # Not compliant with the HTML spec, but backward-compatible
                    # cf. https://github.com/py-pdf/fpdf2/pull/1217#discussion_r1666643777
                    self.follows_heading = True
                if isinstance(tag_style, TextStyle):
                    t_margin = tag_style.t_margin
                    b_margin = tag_style.b_margin
                    l_margin = tag_style.l_margin
                else:
                    t_margin = b_margin = l_margin = 0.0
                l_margin = self._normalize_l_margin(l_margin)
                self._new_paragraph(
                    line_height=(
                        self.line_height_stack[-1] if self.line_height_stack else None
                    ),
                    top_margin=t_margin,
                    bottom_margin=b_margin,
                    indent=l_margin,
                )
        if tag == "ul":
            self.indent += 1
            bullet_char = attrs_dict.get("type", self.ul_bullet_char)
            assert bullet_char is not None
            self.bullet.append(bullet_char)
            line_height_str = css_style.get(
                "line-height", attrs_dict.get("line-height")
            )
            # "line-height" attributes are not valid in HTML,
            # but we support it for backward compatibility,
            # because fpdf2 honors it since 2.6.1 and PR #629
            if line_height_str:
                try:
                    # YYY parse and convert non-float line_height values
                    self.line_height_stack.append(float(line_height_str))
                except ValueError:
                    pass
            else:
                self.line_height_stack.append(None)
            if self.indent == 1:
                tag_style = self.tag_styles[tag]
                if isinstance(tag_style, TextStyle):
                    t_margin = tag_style.t_margin
                    b_margin = tag_style.b_margin
                    l_margin = tag_style.l_margin
                else:
                    t_margin = b_margin = l_margin = 0.0
                l_margin = self._normalize_l_margin(l_margin)
                self._new_paragraph(
                    line_height=0,
                    top_margin=t_margin,
                    bottom_margin=b_margin,
                    indent=l_margin,
                )
                self._write_paragraph("\u00a0")
            self._end_paragraph()
        if tag == "ol":
            self.indent += 1
            start = 1
            if "start" in attrs_dict:
                start_str = attrs_dict["start"] or "1"
                try:
                    start = int(start_str)
                except ValueError:
                    start = 1
            self.bullet.append(start - 1)
            self.ol_type[self.indent] = attrs_dict.get("type") or "1"
            line_height_str = css_style.get("line-height") or attrs_dict.get(
                "line-height"
            )
            # "line-height" attributes are not valid in HTML,
            # but we support it for backward compatibility,
            # because fpdf2 honors it since 2.6.1 and PR #629
            if line_height_str:
                try:
                    # YYY parse and convert non-float line_height values
                    self.line_height_stack.append(float(line_height_str))
                except ValueError:
                    pass
            else:
                self.line_height_stack.append(None)
            if self.indent == 1:
                tag_style = self.tag_styles[tag]
                if isinstance(tag_style, TextStyle):
                    t_margin = tag_style.t_margin
                    b_margin = tag_style.b_margin
                    l_margin = tag_style.l_margin
                else:
                    t_margin = b_margin = l_margin = 0.0
                l_margin = self._normalize_l_margin(l_margin)
                self._new_paragraph(
                    line_height=0,
                    top_margin=t_margin,
                    bottom_margin=b_margin,
                    indent=l_margin,
                )
                self._write_paragraph("\u00a0")
            self._end_paragraph()
        if tag == "li":
            prev_text_color = self.pdf.text_color
            self.pdf.text_color = self.li_prefix_color
            if self.bullet:
                bullet = self.bullet[self.indent - 1]
            else:
                # Allow <li> to be used outside of <ul> or <ol>.
                bullet = self.ul_bullet_char
            if not isinstance(bullet, int):
                bullet = ul_prefix(bullet, self.pdf.is_ttf_font)
            if not isinstance(bullet, str):
                bullet += 1
                self.bullet[self.indent - 1] = bullet
                ol_type = self.ol_type[self.indent]
                bullet = f"{ol_prefix(ol_type, bullet)}."
            tag_style = self.tag_styles[tag]
            if isinstance(tag_style, TextStyle):
                b_margin = tag_style.b_margin
                l_margin = tag_style.l_margin
                t_margin = tag_style.t_margin
            else:
                b_margin = l_margin = t_margin = 0.0
            l_margin = self._normalize_l_margin(l_margin)
            self._ln(t_margin)
            numeric_indent = 0.0 if isinstance(l_margin, Align) else l_margin
            self._new_paragraph(
                line_height=(
                    self.line_height_stack[-1] if self.line_height_stack else None
                ),
                indent=numeric_indent * self.indent,
                bottom_margin=b_margin,
                bullet=bullet,
            )
            self.pdf.text_color = prev_text_color
        if tag == "font":
            self.style_stack.append(
                FontFace(
                    family=self.font_family,
                    emphasis=self.font_emphasis,
                    size_pt=self.font_size_pt,
                    color=self.font_color,
                )
            )
            if "color" in attrs_dict:
                self.font_color = color_as_decimal(attrs_dict["color"])
            if "font-size" in css_style:
                font_size_str = css_style.get("font-size") or ""
                try:
                    self.font_size_pt = float(font_size_str)
                except ValueError:
                    pass
            elif "size" in attrs_dict:
                font_size_str = attrs_dict.get("size") or ""
                try:
                    self.font_size_pt = float(font_size_str)
                except ValueError:
                    pass
            if "face" in attrs_dict:
                font_family = attrs_dict.get("face") or ""
                self.font_family = font_family.lower()
        if tag == "table":
            self._end_paragraph()
            width: Optional[float] = None
            width_str = css_style.get("width") or attrs_dict.get("width")
            if width_str:
                if width_str[-1] == "%":
                    width = self.pdf.epw * float(width_str[:-1]) / 100
                else:
                    width = float(width_str) / self.pdf.k
            if "border" not in attrs_dict:  # default borders
                borders_layout = (
                    "HORIZONTAL_LINES"
                    if self.table_line_separators
                    else "SINGLE_TOP_LINE"
                )
            else:
                borders_str = attrs_dict["border"] or ""
                try:
                    if int(borders_str) > 0:  # explicitly enabled borders
                        borders_layout = (
                            "ALL"
                            if self.table_line_separators
                            else "NO_HORIZONTAL_LINES"
                        )
                    else:  # explicitly disabled borders
                        borders_layout = "NONE"
                except ValueError:
                    borders_layout = "NONE"
            align = Align.coerce(attrs_dict.get("align") or "CENTER")
            padding = (
                float(attrs_dict["cellpadding"] or 0)
                if "cellpadding" in attrs_dict
                else None
            )
            spacing = float(attrs_dict.get("cellspacing") or 0)
            self.table = Table(
                self.pdf,
                align=align,
                borders_layout=borders_layout,
                line_height=self.font_size_pt / self.pdf.k * self.TABLE_LINE_HEIGHT,
                width=width,
                padding=padding,
                gutter_width=spacing,
                gutter_height=spacing,
            )
            self._ln()
        if tag == "tr":
            if not self.table:
                raise FPDFException("Invalid HTML: <tr> used outside any <table>")
            self.tr = {k.lower(): v for k, v in attrs_dict.items()}
            self.table_row = self.table.row()
        if tag in ("td", "th"):
            if not self.table or not self.table_row:
                raise FPDFException(f"Invalid HTML: <{tag}> used outside any <tr>")
            self.td_th = {k.lower(): v for k, v in attrs_dict.items()}
            self.td_th["tag"] = tag
            if tag == "th":
                if "align" not in self.td_th:
                    self.td_th["align"] = "CENTER"
                self.td_th["b"] = True
            elif len(self.table.rows) == 1 and not self.table_row.cells:
                # => we are in the 1st <tr>, and the 1st cell is a <td>
                # => we do not treat the first row as a header
                # pylint: disable=protected-access
                self.table._first_row_as_headings = (
                    False  # pyright: ignore[reportPrivateUsage]
                )
                self.table._num_heading_rows = 0  # pyright: ignore[reportPrivateUsage]
            if "height" in attrs_dict:
                LOGGER.warning(
                    'Ignoring unsupported height="%s" specified on a <%s>',
                    attrs_dict["height"],
                    tag,
                )
            if "width" in attrs_dict:
                width_str = attrs_dict["width"] or "0"
                # pylint: disable=protected-access
                if len(self.table.rows) == 1:  # => first table row
                    if width_str[-1] == "%":
                        width = float(width_str[:-1])
                    else:
                        width = float(width_str)
                    if not self.table._col_widths:
                        self.table._col_widths = []
                    assert isinstance(self.table._col_widths, list)
                    self.table._col_widths.append(width)
                else:
                    LOGGER.warning(
                        'Ignoring width="%s" specified on a <%s> that is not in the first <tr>',
                        width_str,
                        tag,
                    )
        if tag == "img" and "src" in attrs_dict:
            width = float(attrs_dict.get("width") or 0) / self.pdf.k
            height = float(attrs_dict.get("height") or 0) / self.pdf.k
            if self.table_row:  # => <img> in a <table>
                if width or height:
                    LOGGER.warning(
                        'Ignoring unsupported "width" / "height" set on <img> element'
                    )
                if self.align:
                    LOGGER.warning("Ignoring unsupported <img> alignment")
                self.table_row.cell(img=attrs_dict["src"], img_fill_width=True)
                assert self.td_th is not None
                self.td_th["inserted"] = True
                return
            x: float | Align = self.pdf.get_x()
            if self.align:
                x = self.align
            self.pdf.image(
                self.image_map(attrs_dict["src"] or ""),
                x=x,
                w=width,
                h=height,
                link=self.href,
            )
        if tag == "toc":
            self._end_paragraph()
            self.pdf.insert_toc_placeholder(
                self.render_toc, pages=int(attrs_dict.get("pages") or "1")
            )
        if tag == "sup":
            self.pdf.char_vpos = CharVPos.SUP
        if tag == "sub":
            self.pdf.char_vpos = CharVPos.SUB
        if tag == "title":
            self._in_title = True
        if css_style.get("break-after") == "page":
            if tag in ("br", "hr", "img"):
                self._end_paragraph()
                # pylint: disable=protected-access
                self.pdf._perform_page_break()  # pyright: ignore[reportPrivateUsage]
            else:
                self._page_break_after_paragraph = True

    def handle_endtag(self, tag: str) -> None:
        while (
            self._tags_stack
            and tag != self._tags_stack[-1]
            and self._tags_stack[-1] in self.HTML_UNCLOSED_TAGS
        ):
            self._tags_stack.pop()
        if not self._tags_stack:
            if self.warn_on_tags_not_matching:
                LOGGER.warning(
                    "Unexpected HTML end tag </%s>, start tag may be missing?", tag
                )
        elif tag == self._tags_stack[-1]:
            self._tags_stack.pop()
        elif self.warn_on_tags_not_matching:
            LOGGER.warning(
                "Unexpected HTML end tag </%s>, start tag was <%s>",
                tag,
                self._tags_stack[-1],
            )
        if tag == "a":
            self.href = ""
        if tag == "p":
            if self.style_stack:
                font_face = self.style_stack.pop()
                self.font_family = font_face.family or self.font_family
                self.font_size_pt = font_face.size_pt or self.font_size_pt
                self.font_emphasis = font_face.emphasis or TextEmphasis.NONE
                self.font_color = font_face.color
            self._end_paragraph()
            self.align = None
        if tag in HEADING_TAGS:
            self.heading_level = None
            if self.style_stack:
                font_face = self.style_stack.pop()
                self.font_family = font_face.family or self.font_family
                self.font_size_pt = font_face.size_pt or self.font_size_pt
                self.font_emphasis = font_face.emphasis or TextEmphasis.NONE
                self.font_color = font_face.color
            self._end_paragraph()
            self.follows_heading = True  # We don't want extra space below a heading.
        if tag in (
            "b",
            "blockquote",
            "center",
            "code",
            "em",
            "i",
            "dd",
            "dt",
            "pre",
            "s",
            "strong",
            "u",
        ):
            if self.style_stack:
                font_face = self.style_stack.pop()
                self.font_family = font_face.family or self.font_family
                self.font_size_pt = font_face.size_pt or self.font_size_pt
                self.font_emphasis = font_face.emphasis or TextEmphasis.NONE
                self.font_color = font_face.color
            if tag == "pre":
                self._pre_formatted = False
                self._pre_started = False
            if tag in BLOCK_TAGS:
                self._end_paragraph()
        if tag in ("ul", "ol"):
            self._end_paragraph()
            if tag == "ol":
                self.ol_type.pop(self.indent)
            self.indent -= 1
            self.line_height_stack.pop()
            self.bullet.pop()
        if tag == "table":
            assert self.table is not None
            self.table.render()
            self.table = None
            self._ln()
        if tag == "tr":
            self.tr = None
            self.table_row = None
        if tag in ("td", "th"):
            assert (
                self.td_th is not None
                and self.tr is not None
                and self.table_row is not None
            )
            if "inserted" not in self.td_th:
                # handle_data() was not called => we call it to produce an empty cell:
                bgcolor = color_as_decimal(
                    self.td_th.get("bgcolor") or self.tr.get("bgcolor", None)
                )
                style = FontFace(fill_color=bgcolor) if bgcolor else None
                colspan = int(self.td_th.get("colspan", "1"))
                rowspan = int(self.td_th.get("rowspan", "1"))
                self.table_row.cell(
                    text="", style=style, colspan=colspan, rowspan=rowspan
                )
            self.td_th = None
        if tag == "font":
            if self.style_stack:
                font_face = self.style_stack.pop()
                self.font_family = font_face.family or self.font_family
                self.font_size_pt = font_face.size_pt or self.font_size_pt
                self.font_emphasis = font_face.emphasis or TextEmphasis.NONE
                self.font_color = font_face.color
        if tag == "sup":
            self.pdf.char_vpos = CharVPos.LINE
        if tag == "sub":
            self.pdf.char_vpos = CharVPos.LINE
        if tag == "title":
            self._in_title = False

    def feed(self, data: str) -> None:
        super().feed(data)
        while self._tags_stack and self._tags_stack[-1] in self.HTML_UNCLOSED_TAGS:
            self._tags_stack.pop()
        self._end_paragraph()  # render the final chunk of text and clean up our local context.
        if self._tags_stack and self.warn_on_tags_not_matching:
            LOGGER.warning("Missing HTML end tag for <%s>", self._tags_stack[-1])

    def put_link(self, text: str) -> None:
        "Insert a hyperlink"
        prev_style = FontFace(
            family=self.font_family,
            emphasis=self.font_emphasis,
            size_pt=self.font_size_pt,
            color=self.font_color,
        )
        tag_style = self.tag_styles["a"]
        if tag_style.color:
            self.font_color = tag_style.color
        self.font_family = tag_style.family or self.font_family
        self.font_size_pt = tag_style.size_pt or self.font_size_pt
        if tag_style.emphasis:
            self.font_emphasis |= tag_style.emphasis
        self._write_paragraph(text, link=self.href)
        # Restore previous style:
        self.font_family = prev_style.family or self.font_family
        self.font_size_pt = prev_style.size_pt or self.font_size_pt
        self.font_emphasis = prev_style.emphasis or TextEmphasis.NONE
        self.font_color = prev_style.color

    # pylint: disable=no-self-use
    def render_toc(self, pdf: "FPDF", outline: list[OutlineSection]) -> None:
        "This method can be overridden by subclasses to customize the Table of Contents style."
        pdf.ln()
        for section in outline:
            link = pdf.add_link(page=section.page_number)
            text = f'{" " * section.level * 2} {section.name}'
            text += f' {"." * (60 - section.level*2 - len(section.name))} {section.page_number}'
            pdf.multi_cell(
                w=pdf.epw,
                h=pdf.font_size,
                text=text,
                new_x=XPos.LMARGIN,
                new_y=YPos.NEXT,
                link=link,
            )

    # Subclasses of _markupbase.ParserBase must implement this:
    def error(self, message: str) -> None:
        raise RuntimeError(message)


def _scale_units(
    pdf: "FPDF", in_tag_styles: dict[str, FontFace | TextStyle]
) -> dict[str, FontFace | TextStyle]:
    conversion_factor = get_scale_factor("mm") / pdf.k
    out_tag_styles: dict[str, FontFace | TextStyle] = {}
    for tag_name, tag_style in in_tag_styles.items():
        if isinstance(tag_style, TextStyle):
            out_tag_styles[tag_name] = tag_style.replace(
                t_margin=tag_style.t_margin * conversion_factor,
                l_margin=(
                    tag_style.l_margin * conversion_factor
                    if isinstance(tag_style.l_margin, (int, float))
                    else tag_style.l_margin
                ),
                b_margin=tag_style.b_margin * conversion_factor,
            )
        else:
            out_tag_styles[tag_name] = tag_style
    return out_tag_styles


def ul_prefix(ul_type: str, is_ttf_font: bool) -> str:
    if ul_type == "disc":
        return BULLET_UNICODE if is_ttf_font else MESSAGE_WAITING_WIN1252
    if ul_type == "circle":
        return RING_OPERATOR_UNICODE if is_ttf_font else DEGREE_SIGN_WIN1252
    if len(ul_type) == 1:
        return ul_type
    raise NotImplementedError(f"Unsupported type: {ul_type}")


def ol_prefix(ol_type: str, index: int) -> int | str:
    if ol_type == "1":
        return index
    if ol_type == "a":
        return ascii_lowercase[index - 1]
    if ol_type == "A":
        return ascii_uppercase[index - 1]
    if ol_type == "I":
        return int2roman(index)
    if ol_type == "i":
        return int2roman(index).lower()
    raise NotImplementedError(f"Unsupported type: {ol_type}")


class HTMLMixin:
    """
    [**DEPRECATED since v2.6.0**]
    You can now directly use the `FPDF.write_html()` method
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        warnings.warn(
            (
                "The HTMLMixin class is deprecated since v2.6.0. "
                "Simply use the FPDF class as a replacement."
            ),
            DeprecationWarning,
            stacklevel=get_stack_level(),
        )



from fpdf.structure_tree import PDFObject, StructureTreeBuilder


def test_pdf_object_serialize():
    class Point(PDFObject):
        __slots__ = ("_id", "x", "y")

        def __init__(self, x=0, y=0):
            super().__init__()
            self.x = x
            self.y = y

    class Square(PDFObject):
        __slots__ = ("_id", "top_left", "bottom_right")

        def __init__(self, top_left, bottom_right):
            super().__init__()
            self.top_left = top_left
            self.bottom_right = bottom_right

    point_a = Point()
    point_b = Point(x=10, y=10)
    square = Square(top_left=point_a, bottom_right=point_b)
    point_a.id = 1
    point_b.id = 2
    square.id = 3
    pdf_content = (
        point_a.serialize() + "\n" + point_b.serialize() + "\n" + square.serialize()
    )
    assert pdf_content == """\
1 0 obj
<<
/X 0
/Y 0
>>
endobj
2 0 obj
<<
/X 10
/Y 10
>>
endobj
3 0 obj
<<
/BottomRight 2 0 R
/TopLeft 1 0 R
>>
endobj"""


def _serialize(struct_builder, first_object_id=1):
    n = first_object_id
    for obj in struct_builder:
        obj.id = n
        n += 1
    return "\n".join(obj.serialize() for obj in struct_builder)


def test_empty_structure_tree():
    struct_builder = StructureTreeBuilder()
    assert _serialize(struct_builder) == """\
1 0 obj
<<
/K [2 0 R]
/ParentTree 3 0 R
/Type /StructTreeRoot
>>
endobj
2 0 obj
<<
/K []
/P 1 0 R
/S /Document
/Type /StructElem
>>
endobj
3 0 obj
<<
/Nums []
>>
endobj"""


def test_single_image_structure_tree():
    struct_builder = StructureTreeBuilder()
    struct_builder.add_marked_content(
        1, "/Figure", 0, "Image title", "Image description"
    )
    assert _serialize(struct_builder, first_object_id=3) == """\
3 0 obj
<<
/K [4 0 R]
/ParentTree 5 0 R
/Type /StructTreeRoot
>>
endobj
4 0 obj
<<
/K [6 0 R]
/P 3 0 R
/S /Document
/Type /StructElem
>>
endobj
5 0 obj
<<
/Nums [0 [6 0 R]]
>>
endobj
6 0 obj
<<
/Alt (Image description)
/K [0]
/P 4 0 R
/S /Figure
/T (Image title)
/Type /StructElem
>>
endobj"""



from fpdf import FPDF
from pystrich.datamatrix import DataMatrixEncoder, DataMatrixRenderer


class PDF(FPDF):
    def datamatrix(self, text, w, h=None, x=None, y=None, cellsize=5):
        if x is None:
            x = self.x
        if y is None:
            y = self.y
        if h is None:
            h = w
        encoder = DataMatrixEncoder(text)
        encoder.width = w
        encoder.height = h
        renderer = DataMatrixRenderer(encoder.matrix, encoder.regions)
        img = renderer.get_pilimage(cellsize)
        self.image(img, x, y, w, h)


# Usage example:
pdf = PDF()
pdf.add_page()
pdf.set_font("Helvetica", size=24)
pdf.datamatrix("Hello world!", w=100)
pdf.output("datamatrix_from_method.pdf")



from pathlib import Path

from fpdf import FPDF
from fpdf.enums import TableCellFillMode, VAlign
from fpdf.fonts import FontFace
from fpdf.table import draw_box_borders

from test.conftest import assert_pdf_equal, LOREM_IPSUM

HERE = Path(__file__).resolve().parent


IMG_DIR = HERE.parent / "image"

IMAGE = IMG_DIR / "image_types/pythonknight.png"

IMAGES_DATA = (
    (IMAGE, IMAGE, IMAGE, IMAGE),
    (IMAGE, IMAGE, IMAGE, IMAGE),
    (IMAGE, IMAGE, IMAGE, IMAGE),
    (IMAGE, IMAGE, IMAGE, IMAGE),
    (IMAGE, IMAGE, IMAGE, IMAGE),
    (IMAGE, IMAGE, IMAGE, IMAGE),
    (IMAGE, IMAGE, IMAGE, IMAGE),
)

MULTILINE_TABLE_DATA = (
    ("Multilines text", "Image"),
    (LOREM_IPSUM[:200], IMG_DIR / "png_images/ba2b2b6e72ca0e4683bb640e2d5572f8.png"),
    (LOREM_IPSUM[200:400], IMG_DIR / "png_images/ac6343a98f8edabfcc6e536dd75aacb0.png"),
    (LOREM_IPSUM[400:600], IMG_DIR / "image_types/insert_images_insert_png.png"),
    (LOREM_IPSUM[600:800], IMG_DIR / "image_types/circle.bmp"),
)

TABLE_DATA = (
    ("First name", "Last name", "Age", "City"),
    ("Jules", "Smith", "34", "San Juan"),
    (
        "Brave Sir Robin ",
        "the  Not-Quite-So-Brave-As-Sir-Lancelot",
        "+",
        "Bridge of Death",
    ),
    ("Mary", "Ramos", "45", "Orlando"),
    ("Carlson", "Banks", "19", "Los Angeles"),
    ("Lucas", "Cimon", "31", "Angers"),
)

LONG_TEXT = (
    "\nProfessor: (Eric Idle) It's an entirely new strain of sheep, a killer sheep that can not only hold a rifle but is also a first-class shot.\n"
    "Assistant: But where are they coming from, professor?\n"
    "Professor: That I don't know. I just don't know. I really just don't know. I'm afraid I really just don't know. I'm afraid even I really just"
    " don't know. I have to tell you I'm afraid even I really just don't know. I'm afraid I have to tell you... (she hands him a glass of water"
    " which she had been busy getting as soon as he started into this speech) ... thank you ... (resuming normal breezy voice) ... I don't know."
    " Our only clue is this portion of wolf's clothing which the killer sheep ..."
)


TABLE_DATA_LIST = ["And", "now", "for", "something", "completely", "different"]


def test_table_with_multiline_cells_and_images_padding_and_pagebreak(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=16)

    red = (255, 100, 100)
    black = (0, 0, 0)

    deathstyle = FontFace(color=black, fill_color=red)

    with pdf.table(
        line_height=pdf.font_size,
        padding=(5, 5, 5, 5),
        col_widths=(0.3, 0.2),
        width=80,
        outer_border_width=3,
    ) as table:
        for i, data_row in enumerate(MULTILINE_TABLE_DATA):
            row = table.row()
            for j, datum in enumerate(data_row):
                if j == 1 and i > 0:
                    row.cell(img=datum, img_fill_width=False, style=deathstyle)
                else:
                    row.cell(datum)

    assert_pdf_equal(
        pdf,
        HERE / "table_with_multiline_cells_and_images_padding_and_pagebreak.pdf",
        tmp_path,
    )


def test_table_with_single_row_of_images(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=16)
    pdf.text(60, 100, "All these images should be the same size")
    with pdf.table(
        line_height=pdf.font_size,
        padding=(5, 7, 3, 4),
        width=120,
        col_widths=(2, 2, 2, 2),
        cell_fill_color=(150, 200, 255),
        cell_fill_mode=TableCellFillMode.ROWS,
    ) as table:
        data_row = IMAGES_DATA[0]
        row = table.row()
        for datum in data_row:
            row.cell(img=datum)

    assert_pdf_equal(pdf, HERE / "table_with_single_row_of_images.pdf", tmp_path)


def test_table_with_only_images(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=16)
    with pdf.table(
        line_height=pdf.font_size,
        padding=(5, 7, 3, 4),
        width=120,
        col_widths=(1, 2, 3, 2),
        cell_fill_color=(150, 200, 255),
        cell_fill_mode=TableCellFillMode.ROWS,
    ) as table:
        for data_row in IMAGES_DATA:
            row = table.row()
            for datum in data_row:
                row.cell(img=datum)

    assert_pdf_equal(pdf, HERE / "table_with_only_images.pdf", tmp_path)


def test_table_vertical_alignment(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=12)

    red = (255, 100, 100)
    black = (0, 0, 0)

    pdf.rect(pdf.l_margin, pdf.t_margin, pdf.epw, pdf.eph)
    pdf.rect(0, pdf.t_margin, pdf.l_margin, pdf.eph)
    pdf.rect(pdf.l_margin + pdf.epw, pdf.t_margin, pdf.r_margin, pdf.eph)
    # pdf.rect(0, pdf.t_margin, pdf.l_margin, pdf.eph)

    deathstyle = FontFace(color=black, fill_color=red)
    # write_html() doesn't add a nl at the very top of the page anymore.
    pdf.ln(0.1)

    for v in (VAlign.T, VAlign.M, VAlign.B):
        pdf.write_html(f"<h1>Vertical alignment: {v}</h1>")

        with pdf.table(line_height=pdf.font_size, padding=3, v_align=v) as table:
            for data_row in TABLE_DATA:
                row = table.row()
                for datum in data_row:
                    if "Death" in datum:
                        row.cell(datum, style=deathstyle)
                    else:
                        row.cell(datum)

    assert_pdf_equal(pdf, HERE / "table_vertical_alignment.pdf", tmp_path)


def test_table_padding_per_cell(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=12)

    red = (255, 100, 100)
    black = (0, 0, 0)

    deathstyle = FontFace(color=black, fill_color=red)
    with pdf.table(line_height=pdf.font_size, padding=2) as table:
        for irow in range(5):
            row = table.row()
            for icol in range(5):
                datum = "Circus"
                if irow == 3 and icol % 2 == 0:
                    row.cell(
                        "custom padding", style=deathstyle, padding=(2 * icol, 8, 8, 8)
                    )
                else:
                    row.cell(datum)

    assert_pdf_equal(pdf, HERE / "table_padding_per_cell.pdf", tmp_path)


def test_table_valign_per_row(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=12)
    with pdf.table(first_row_as_headings=False, padding=2) as table:
        for _irow, v_align in enumerate(VAlign):
            row = table.row(v_align=v_align)
            for icol in range(5):
                datum = icol * "Circus\n"
                row.cell(
                    f"{datum}v_align={v_align}",
                    v_align=v_align,
                )
    assert_pdf_equal(pdf, HERE / "table_valign_per_row.pdf", tmp_path)


def test_table_valign_per_cell(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=12)

    red = (255, 100, 100)
    black = (0, 0, 0)

    deathstyle = FontFace(color=black, fill_color=red)

    with pdf.table(line_height=pdf.font_size, padding=2, v_align=VAlign.M) as table:
        for irow in range(5):
            row = table.row()
            for icol in range(5):
                datum = icol * "Circus\n"
                if irow in (2, 3):
                    if irow == 2:
                        v_align = VAlign.T
                    else:
                        v_align = VAlign.B
                    row.cell(
                        f"{datum}: custom v-align {v_align}",
                        style=deathstyle,
                        v_align=v_align,
                    )
                else:
                    row.cell(datum)

    assert_pdf_equal(pdf, HERE / "table_valign_per_cell.pdf", tmp_path)


def test_table_with_gutter_and_padding_and_outer_border_width(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=16)
    with pdf.table(
        TABLE_DATA,
        gutter_height=3,
        gutter_width=5,
        line_height=pdf.font_size,
        align="L",
        outer_border_width=2,
    ):
        pass

    assert_pdf_equal(
        pdf, HERE / "table_with_gutter_and_padding_and_outer_border_width.pdf", tmp_path
    )


def test_table_with_colspan(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=12)

    def make_text(n):
        a = TABLE_DATA_LIST[int(n / 2) % len(TABLE_DATA_LIST)]
        b = TABLE_DATA_LIST[(int(n**2 * 10) + 4) % len(TABLE_DATA_LIST)]

        if n % 4 == 0:
            return f"{a} {b}"
        if n % 4 == 1:
            return f"{a} {b} {a}"
        if n % 4 == 2:
            return f"{b} {a}"

        return a

    with pdf.table(line_height=pdf.font_size, gutter_height=3, gutter_width=3) as table:
        for irow in range(10):
            row = table.row()
            for icol in range(10):
                txt = make_text(irow * 10 + icol)
                txt = str(icol) + " " + txt
                if irow > 0:
                    if icol == 2:
                        row.cell(txt + " SPAN 2", colspan=2)
                        continue
                    if icol == 3:
                        continue
                row.cell(txt)

    assert_pdf_equal(pdf, HERE / "table_with_colspan.pdf", tmp_path)


def test_table_outside_border_width(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=12)

    with pdf.table(outer_border_width=1, gutter_height=10, gutter_width=15) as table:
        for _ in range(5):
            row = table.row()
            for _ in range(5):
                datum = "Circus"
                row.cell(datum)

    assert_pdf_equal(pdf, HERE / "table_with_outside_border_width.pdf", tmp_path)


def test_table_colspan_and_padding(tmp_path):
    pdf = FPDF()

    pdf.set_font("Times", size=12)
    pdf.add_page()
    with pdf.table(col_widths=(1, 2, 1, 1), padding=5) as table:
        row = table.row()
        row.cell("0")
        row.cell("1")
        row.cell("2")
        row.cell("3")
        row = table.row()
        row.cell("A1")
        row.cell("A2", colspan=2)
        # row.cell("void") # <--- this cell is not rendered
        row.cell("A4")

        row = table.row()
        row.cell("B1", colspan=2)
        # row.cell("void")  # <--- this cell is not rendered
        row.cell("B3")
        row.cell("B4")

    pdf.c_margin = 10

    with pdf.table(col_widths=(1, 2, 1, 1), padding=5) as table:
        row = table.row()
        row.cell("0")
        row.cell("1")
        row.cell("2")
        row.cell("3")
        row = table.row()
        row.cell("A1")
        row.cell("A2", colspan=2)
        # row.cell("void")  # <--- this cell is not rendered
        row.cell("A4")

        row = table.row()
        row.cell("B1", colspan=2)
        # row.cell("void")  # <--- this cell is not rendered
        row.cell("B3")
        row.cell("B4")

    assert_pdf_equal(pdf, HERE / "table_colspan_and_padding.pdf", tmp_path)


def test_table_colspan_and_padding_and_gutter(tmp_path):
    pdf = FPDF()

    pdf.set_font("Times", size=12)
    pdf.add_page()
    with pdf.table(
        col_widths=(1, 2, 1, 1), padding=5, gutter_height=3, gutter_width=5
    ) as table:
        row = table.row()
        row.cell("0")
        row.cell("1")
        row.cell("2")
        row.cell("3")
        row = table.row()
        row.cell("A1")
        row.cell("A2", colspan=2)
        row.cell("A4")

        row = table.row()
        row.cell("B1", colspan=2)
        row.cell("B3")
        row.cell("B4")

    assert_pdf_equal(pdf, HERE / "table_colspan_and_padding_and_gutter.pdf", tmp_path)


def test_table_colspan_and_padding_and_gutter_and_width(tmp_path):
    pdf = FPDF()

    pdf.set_font("Times", size=12)
    pdf.add_page()
    with pdf.table(
        col_widths=(1, 2, 1, 1), padding=5, gutter_height=3, gutter_width=5, width=100
    ) as table:
        row = table.row()
        row.cell("0")
        row.cell("1")
        row.cell("2")
        row.cell("3")
        row = table.row()
        row.cell("A1")
        row.cell("A2,3,4", colspan=3)

        row = table.row()
        row.cell("B1", colspan=2)
        row.cell("B3")
        row.cell("B4")

    assert_pdf_equal(
        pdf, HERE / "table_colspan_and_padding_and_gutter_and_width.pdf", tmp_path
    )


def test_table_with_cell_overflow(tmp_path):
    pdf = FPDF()
    pdf.set_font("Times", size=30)
    pdf.add_page()

    boldstyle = FontFace("Times", emphasis="B")

    with pdf.table(width=pdf.epw / 2, col_widths=(1, 2, 1)) as table:
        row = table.row()
        row.cell("left")
        row.cell("center")
        row.cell("right")  # triggers header cell overflow on last column
        row = table.row()
        row.cell("left")
        row.cell("center")
        row.cell(
            "right", style=boldstyle
        )  # triggers header cell overflow on last column
        row = table.row()
        row.cell("A1")
        row.cell("A2")
        row.cell("A33333333")  # triggers cell overflow on last column
        row = table.row()
        row.cell("B1")
        row.cell("B2")
        row.cell("B3")

    assert_pdf_equal(pdf, HERE / "table_with_cell_overflow_font_setting.pdf", tmp_path)


def test_table_draw_box_borders(tmp_path):
    pdf = FPDF()
    pdf.set_font("Times", size=16)
    pdf.add_page()

    def box(x, y, borders):
        draw_box_borders(
            pdf,
            x - 10,
            y - 10,
            x + 40,
            y + 20,
            fill_color=(200, 200, 200),
            border="None",
        )
        draw_box_borders(pdf, x - 20, y - 20, x + 50, y + 30, border=borders)
        pdf.set_xy(x, y)
        pdf.cell(text=borders)

    box(40, 40, "L")
    box(140, 40, "R")

    box(40, 140, "T")
    box(140, 140, "B")

    assert_pdf_equal(pdf, HERE / "table_draw_box_borders.pdf", tmp_path)



from pathlib import Path

import pytest

from fpdf import FPDF
from test.conftest import assert_pdf_equal

HERE = Path(__file__).resolve().parent


@pytest.mark.xfail(
    reason="Implementation not finished, cf. https://github.com/py-pdf/fpdf2/issues/62"
)
def test_linearization(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.image(
        HERE / "image/png_images/66ac49ef3f48ac9482049e1ab57a53e9.png", x=150, y=150
    )
    assert_pdf_equal(pdf, HERE / "linearization.pdf", tmp_path, linearize=True)



from pathlib import Path

from fpdf import FPDF
from test.conftest import LOREM_IPSUM, assert_pdf_equal

HERE = Path(__file__).resolve().parent
FONTS_DIR = HERE.parent / "fonts"


def test_sbix_compyx(tmp_path):
    # Compyx - OFL license
    # https://github.com/RoelN/Compyx

    pdf = FPDF()
    pdf.add_font("Compyx", "", HERE / "Compyx-Regular-SBIX.ttf")

    pdf.add_page()
    pdf.set_font("Compyx", size=16)
    pdf.multi_cell(w=pdf.epw, text=LOREM_IPSUM.lower(), align="L")
    pdf.ln()
    pdf.multi_cell(w=pdf.epw, text=LOREM_IPSUM.lower(), align="R")
    pdf.ln()
    pdf.multi_cell(w=pdf.epw, text=LOREM_IPSUM.lower(), align="J")

    assert_pdf_equal(pdf, HERE / "sbix_compyx.pdf", tmp_path)


def test_sbix_bungee(tmp_path):
    # Bungee Color - OFL license
    # https://github.com/djrrb/Bungee

    pdf = FPDF()
    pdf.add_font("Bungee", "", HERE / "BungeeColor-Regular_sbix_MacOS.ttf")

    pdf.add_page()
    pdf.set_font("Bungee", size=16)
    pdf.multi_cell(w=pdf.epw, text=LOREM_IPSUM.upper(), align="L")
    pdf.ln()
    pdf.multi_cell(w=pdf.epw, text=LOREM_IPSUM.upper(), align="R")
    pdf.ln()
    pdf.multi_cell(w=pdf.epw, text=LOREM_IPSUM.upper(), align="J")

    assert_pdf_equal(pdf, HERE / "sbix_bungee.pdf", tmp_path)



from pathlib import Path

from fpdf import FPDF
from test.conftest import assert_pdf_equal

HERE = Path(__file__).resolve().parent


def test_set_no_margin(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_margin(0)
    pdf.image(HERE / "../image/image_types/insert_images_insert_png.png", w=pdf.epw)
    assert_pdf_equal(pdf, HERE / "set_no_margin.pdf", tmp_path)



from mistletoe import markdown

html = markdown("""
# Unicode:

| Emoji | Description |
| --- | - |
| 😀 | GRINNING FACE |
| 😁 | GRINNING FACE WITH SMILING EYES |
| 😈 | SMILING FACE WITH HORNS |

# A checklist:

* ☐ item 1
* ☑ item 2
* ☐ item 3
""")

from fpdf import FPDF

pdf = FPDF()
pdf.add_font("DejaVuSans", fname="test/fonts/DejaVuSans.ttf")
pdf.add_font("DejaVuSans", fname="test/fonts/DejaVuSans-Bold.ttf", style="B")
pdf.set_font("DejaVuSans", size=24)
pdf.add_page()
pdf.write_html(html)
pdf.output("pdf-from-markdown-with-mistletoe-unicode.pdf")
